import numpy as np

from njet.jet import factorials, check_zero
from njet import derive, jetpoly

from .genfunc import genexp
from .magnus import hard_edge, hard_edge_chain, norsett_iserles

class liepoly:
    '''
    Class to model the Lie operator :p:, where p is a polynomial given in terms of
    complex (xi, eta)-coordinates. For the notation of these coordinates see Ref.
    "M. Titze: Space Charge Modeling at the Integer Resonance for the CERN PS and SPS" (2019),
    p. 33 onwards.
    
    Parameters
    ----------
    self.max_power: int
        A value > 0 means that any calculations leading to expressions beyond this 
        degree will be discarded. For binary operations the minimum of both 
        max_powers are used.
        
    **kwargs
        Optional arguments passed to self.set_monimial
    '''
    def __init__(self, max_power=float('inf'), **kwargs):
        # self.dim denotes the number of xi (or eta)-factors.
        if 'values' in kwargs.keys():
            self._values = kwargs['values']
        elif 'a' in kwargs.keys() or 'b' in kwargs.keys():
            self.set_monomial(**kwargs)
        else:
            self._values = {}
            
        if len(self._values) == 0:
            self.dim = kwargs.get('dim', 0)
        else:
            self.dim = kwargs.get('dim', len(next(iter(self._values)))//2)
            
        self.set_max_power(max_power)
        
    def set_max_power(self, max_power):
        self.max_power = max_power
        self._values = {k: v for k, v in self.items() if sum(k) <= max_power}
        
    def set_monomial(self, a=[], b=[], value=1, **kwargs):
        dim = max([len(a), len(b)])
        if len(a) < dim:
            a += [0]*(dim - len(a))
        if len(b) < dim:
            b += [0]*(dim - len(b))
        self._values = {tuple(a + b): value}
        
    def maxdeg(self):
        '''
        Obtain the maximal degree of the current Lie polynomial. 
        '''
        if len(self._values) == 0:
            return 0
        else:
            return max([sum(k) for k, v in self.items()])
    
    def mindeg(self):
        '''
        Obtain the minimal degree of the current Lie polynomial. 
        '''
        if len(self._values) == 0:
            return 0
        else:
            return min([sum(k) for k, v in self.items()])
        
    def copy(self):
        new_values = {}
        for k, v in self.items():
            if hasattr(v, 'copy'):
                v = v.copy()
            new_values[k] = v
        return self.__class__(values=new_values, dim=self.dim, max_power=self.max_power)
    
    def extract(self, condition):
        '''
        Extract a Lie polynomial from the current Lie polynomial, based on a condition.
        
        Parameters
        ----------
        condition: callable
            A function which maps a given tuple (an index) to a boolean. For example 'condition = lambda x: sum(x) == k' would
            yield the homogeneous part of the current Lie polynomial (this is realized in 'self.homogeneous_part').
            
        Returns
        -------
        liepoly
            The extracted Lie polynomial.
        '''
        return self.__class__(values={key: value for key, value in self.items() if condition(key)}, dim=self.dim, max_power=self.max_power)
    
    def homogeneous_part(self, k: int):
        '''
        Extract the homogeneous part of order k from the current Lie polynomial.
        
        Parameters
        ----------
        k: int
            The requested order.
            
        Returns
        -------
        liepoly
            The extracted Lie polynomial.
        '''
        return self.extract(condition=lambda x: sum(x) == k)
        
    def __call__(self, z):
        '''
        Evaluate the polynomial at a specific position z.
        
        Parameters
        ----------
        z: subscriptable
            The point at which the polynomial should be evaluated. It is assumed that len(z) == self.dim,
            in which case the components of z are assumed to be xi-values. Otherwise, it is assumed that len(z) == 2*self.dim,
            where z = (xi, eta) denote a set of complex conjugated coordinates.
        '''
        # some consistency check
        if z.__class__.__name__ == self.__class__.__name__:
            raise TypeError(f"Input of type '{self.__class__.__name__}' not supported.") # later on, the getitem method of z is called from 0 onward, which will not behave well for liepoly objects.
        
        # prepare input vector
        if len(z) == self.dim:
            z = [e for e in z] + [e.conjugate() for e in z]
        assert len(z) == 2*self.dim, f'Number of input parameters: {len(z)}, expected: {2*self.dim}'
        
        # compute the occuring powers ahead of evaluation
        z_powers = {}
        j = 0
        for we in zip(*self.keys()):
            z_powers[j] = {k: z[j]**int(k) for k in np.unique(we)} # need to convert k to int, 
            # otherwise we get a conversion to some numpy array if z is not a float (e.g. an njet).
            j += 1
            
        # evaluate polynomial at requested point
        result = 0
        for k, v in self.items():
            prod = 1
            for j in range(self.dim):
                prod *= z_powers[j][k[j]]*z_powers[j + self.dim][k[j + self.dim]]
            result += prod*v # v needs to stay on the right-hand side here, because prod may be a jet class (if we
            # compute the derivative(s) of the Lie polynomial)
        return result
        
    def __add__(self, other):
        if other == 0:
            return self
        add_values = {k: v for k, v in self.items()}
        if self.__class__.__name__ != other.__class__.__name__:
            # Treat other object as constant.
            if not check_zero(other):
                zero_tpl = (0,)*self.dim*2
                new_value = add_values.get(zero_tpl, 0) + other
                if not check_zero(new_value):
                    add_values[zero_tpl] = new_value
                else:
                    _ = add_values.pop(zero_tpl, None)
            max_power = self.max_power
        else:
            assert self.dim == other.dim, f'Dimensions do not agree: {self.dim} != {other.dim}'
            for k, v in other.items():
                new_v = add_values.get(k, 0) + v
                if not check_zero(new_v):
                    add_values[k] = new_v
                else:
                    _ = add_values.pop(k, None)
            max_power = min([self.max_power, other.max_power])
        return self.__class__(values=add_values, dim=self.dim, max_power=max_power)
    
    def __radd__(self, other):
        return self + other
    
    def __neg__(self):
        return self.__class__(values={k: -v for k, v in self.items()}, 
                              dim=self.dim, max_power=self.max_power)
    
    def __sub__(self, other):
        return self + -other

    def __matmul__(self, other):
        return self.poisson(other)
        
    def poisson(self, other):
        '''
        Compute the Poisson-bracket {self, other}
        '''
        if self.__class__.__name__ != other.__class__.__name__:
            raise TypeError(f"unsupported operand type(s) for poisson: '{self.__class__.__name__}' and '{other.__class__.__name__}'.")
        assert self.dim == other.dim, f'Dimensions do not agree: {self.dim} != {other.dim}'
        max_power = min([self.max_power, other.max_power])
        poisson_values = {}
        for t1, v1 in self.items():
            power1 = sum(t1)
            for t2, v2 in other.items():
                power2 = sum(t2)
                if power1 + power2 - 2 > max_power:
                    continue
                a, b = t1[:self.dim], t1[self.dim:]
                c, d = t2[:self.dim], t2[self.dim:]
                for k in range(self.dim):
                    det = a[k]*d[k] - b[k]*c[k]
                    if det == 0:
                        continue
                    new_power = tuple([a[j] + c[j] if j != k else a[j] + c[j] - 1 for j in range(self.dim)] + \
                                [b[j] + d[j] if j != k else b[j] + d[j] - 1 for j in range(self.dim)])
                    new_value = v1*v2*det*-1j + poisson_values.get(new_power, 0)
                    if not check_zero(new_value):
                        poisson_values[new_power] = new_value
                    else:
                        _ = poisson_values.pop(new_power, None)
        return self.__class__(values=poisson_values, dim=self.dim, max_power=max_power)
    
    def __mul__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            assert self.dim == other.dim
            dim2 = 2*self.dim
            max_power = min([self.max_power, other.max_power])
            mult_values = {}
            for t1, v1 in self.items():
                power1 = sum(t1)
                for t2, v2 in other.items():
                    power2 = sum(t2)
                    if power1 + power2 > max_power:
                        continue
                    prod_tpl = tuple([t1[k] + t2[k] for k in range(dim2)])
                    prod_val = mult_values.get(prod_tpl, 0) + v1*v2 # it is assumed that v1 and v2 are both not zero, hence prod_val != 0.
                    if not check_zero(prod_val):
                        mult_values[prod_tpl] = prod_val
                    else:
                        _ = mult_values.pop(prod_tpl, None)
            return self.__class__(values=mult_values, dim=self.dim, max_power=max_power)
        else:
            return self.__class__(values={k: v*other for k, v in self.items() if not check_zero(other)}, 
                                          dim=self.dim, max_power=self.max_power) # need to use v*other; not other*v here: If type(other) = numpy.float64, then it may cause unpredicted results if it stands on the left.
        
    def __rmul__(self, other):
        return self.__mul__(other)
    
    def __truediv__(self, other):
        # implement '/' operator
        if self.__class__.__name__ != other.__class__.__name__:
            # Attention: If other is a NumPy array, there is no check if one of the entries is zero.
            return self.__class__(values={k: v/other for k, v in self.items() if not check_zero(other)}, 
                                          dim=self.dim, max_power=self.max_power)
        else:
            raise NotImplementedError('Division by Lie polynomial not supported.')
        
    
    def __pow__(self, other):
        assert type(other) == int
        assert other >= 0
        if other == 0:
            return self.__class__(values={(0,)*self.dim*2: 1}, 
                                  dim=self.dim, max_power=self.max_power) # N.B. 0**0 := 1

        remainder = other%2
        half = self**(other//2)
        if remainder == 1:
            return self*half*half
        else:
            return half*half
        
    def conjugate(self):
        return self.__class__(values={k[self.dim:] + k[:self.dim]: v.conjugate() for k, v in self.items()},
                              dim=self.dim, max_power=self.max_power)
        
    def __len__(self):
        return len(self._values)
    
    def __eq__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            return self._values == other._values
        else:
            if self.maxdeg() != 0:
                return False
            else:
                return self.get((0, 0), 0) == other
            
    def keys(self):
        return self._values.keys()
    
    def get(self, *args, **kwargs):
        return self._values.get(*args, **kwargs)
    
    def items(self):
        return self._values.items()
    
    def values(self):
        return self._values.values()
    
    def __iter__(self):
        for key in self._values.keys():
            yield self._values[key]
            
    def __getitem__(self, key):
        return self._values[key]
    
    def __setitem__(self, key, other):
        self._values[key] = other
        
    def update(self, d):
        new_values = {k: v for k, v in self.items()}
        new_values.update(d)
        return self.__class__(values=new_values, dim=self.dim, max_power=self.max_power)
        
    def ad(self, y, power: int=1):
        '''
        Compute repeated Poisson-brackets.
        E.g. let x = self. Then {x, {x, {x, {x, {x, {x, y}}}}}} =: x**6(y)
        Special case: x**0(y) := y
        
        Let z be a homogeneous Lie polynomial and deg(z) the degree of z. Then it holds
        deg(x**m(y)) = deg(y) + m*(deg(x) - 2).
        This also holds for the maximal degrees and minimal degrees in case that x, and y
        are inhomogeneous.
        
        Therefore, if x and y both have non-zero max_power fields, 'self.power' will not evaluate
        terms x**m with
        m >= min([(max_power - mindeg(y))//(mindeg(x) - 2), power]).
        
        Parameters
        ----------
        y: liepoly
            Lie polynomial which we want to evaluate on

        power: int, optional
            Number of repeated brackets (default: 1).


        Returns
        -------
        list
            List [x**k(y) for k in range(n + 1)], if n is the power requested.
        '''
        if self.__class__.__name__ != y.__class__.__name__:
            raise TypeError(f"unsupported operand type(s) for adjoint: '{self.__class__.__name__}' on '{y.__class__.__name__}'.")
        assert power >= 0
        
        # Adjust requested power if max_power makes this necessary, see comment above.
        max_power = min([self.max_power, y.max_power])
        mindeg_x = self.mindeg()
        if mindeg_x > 2 and max_power < float('inf'):
            mindeg_y = y.mindeg()
            power = min([(max_power - mindeg_y)//(mindeg_x - 2), power]) # N.B. // works as floor division
            
        result = self.__class__(values={k: v for k, v in y.items()}, 
                                dim=y.dim, max_power=max_power)
        all_results = [result]
        # N.B.: We can not set values = self.values, otherwise result.values will get changed if self.values is changing.
        for k in range(power):
            result = self@result
            if len(result) == 0:
                break
            all_results.append(result)
        return all_results
    
    def __str__(self):
        out = ''
        for k, v in self.items():
            out += f'{k}: {str(v)} '
        if len(out) > 0:
            return out[:-1]
        else:
            return '0'

    def _repr_html_(self):
        return f'<samp>{self.__str__()}</samp>'
    
    def derive(self, **kwargs):
        '''
        Derive the current Lie polynomial.
        
        Parameters
        ----------
        order: int
            The order by which we are going to derive the polynomial.
            
        Returns
        -------
        derive
            A class of type njet.ad.derive with n_args=2*self.dim parameters.
            Note that a function evaluation should be consistent with the fact that 
            the last self.dim entries are the complex conjugate values of the 
            first self.dim entries.
        '''
        return derive(self, n_args=2*self.dim, **kwargs)
    
    def flow(self, power: int, t=1, **kwargs):
        '''
        Let f: R^n -> R be a differentiable function and :x: the current polynomial Lie map. 
        Then this routine will compute the components of M: R^n -> R^n,
        where M is the map satisfying
        exp(:x:) f = f o M

        Note that the degree to which powers are discarded is given by self.max_power.

        Parameters
        ----------
        power: int
            The maximal power up to which exp(:x:) should be evaluated.
        
        t: float, optional
            An additional parameter to model exp(t*:x:) (default: 1). Note that
            this parameter can also be changed in the lieoperator class later.
            
        **kwargs
            Additional arguments are passed to the lieoperator class.

        Returns
        -------
        lieoperator
            Class of type lieoperator, modeling the flow of the current Lie polynomial.
        '''
        return lieoperator(self, generator=genexp(power), t=t, **kwargs)
        
    def construct(self, f, **kwargs):
        '''
        Let :x: represent the current Lie polynomial. Depending on the input,
        this routine will either return the map f(x) or the Lie polynomial :f(x):.
        
        Parameters
        ----------
        f: callable
            A function depending on a single parameter. It needs to be supported by the njet module.
            
        **kwargs
            Additional parameters passed to lie.construct routine.
            
        Returns
        -------
        callable or liepoly
            The output depends on the optional argument 'power'.
            
            If no argument 'power' has been passed, then it will
            be taken from the current value self.max_power.
            
            If power < float('inf'), then the Lie polynomial :f(x): is returned,
            where f has been expanded up to the specified power. If power == float('inf'),
            then the function f(x) is returned.
        '''
        if not 'power' in kwargs.keys():
            kwargs['power'] = self.max_power
        return construct([self], f, **kwargs)
    
    def to_jetpoly(self):
        '''
        Map the current Lie polynomial to an njet jetpoly class.
        
        Returns
        -------
        jetpoly
            A jetpoly class of self.dim*2 variables, representing the current Lie polynomial.
        '''
        # N.B. self.dim corresponds to the number of xi (or eta) variables.
        # Although xi and eta are related by complex conjugation, we need to treat them as being independently,
        # in line with Wirtinger calculus. However, this fact needs to be taken into account when evaluating those polynomials, so
        # a polynomial should be evaluated always at points [z, z.conjugate()] etc.

        constant_key = (0,)*self.dim*2
        jpvalues = {}
        if constant_key in self.keys():
            jpvalues[frozenset([(0, 0)])] = self._values[constant_key]
        for key, v in self.items():
            if sum(key) == 0: # we already dealt with the constant term.
                continue
            jpvalues[frozenset([(j, key[j]) for j in range(self.dim*2) if key[j] != 0])] = v
        return jetpoly(values=jpvalues)
    
    def apply(self, name, cargs={}, *args, **kwargs):
        '''
        Apply a class function of the coefficients of the current Lie-polynomial.
        
        Parameters
        ----------
        name: str
            The name of the class function 'name'.
            
        cargs: dict, optional
            Dictionary of keywords which may depend on self.values.keys(). This means that the keys of
            cargs must correspond to self.values.keys(). The items of cargs correspond to a set of keyworded
            arguments for the class function 'name'.
            
        *args:
            Arguments of the class function 'name'.
            
        **kwargs:
            Keyworded arguments of the class function 'name'.
            
        Returns
        -------
        liepoly
            A Lie-polynomial in which every entry in its values contain the result of the requested class function.
        '''
        if len(cargs) > 0:
            out = {key: getattr(v, name)(*args, **cargs[key]) for key, v in self.items()}
        else:
            out = {key: getattr(v, name)(*args, **kwargs) for key, v in self.items()}
        return self.__class__(values=out, dim=self.dim, max_power=self.max_power)
            
    
def create_coords(dim, cartesian=False, **kwargs):
    '''
    Create a set of complex (xi, eta)-Lie polynomials for a given dimension.
    
    Parameters
    ----------
    dim: int
        The requested dimension.
        
    cartesian: boolean, optional
        If true, create cartesian coordinates q and p instead. 
        
        Note that it holds:
        q = (xi + eta)/sqrt(2)
        p = (xi - eta)/sqrt(2)/1j
        
    **kwargs
        Optional arguments passed to liepoly class.
        
    Returns
    -------
    list
        List of length 2*dim with liepoly entries, corresponding to the xi_k and eta_k Lie polynomials. Hereby the first
        dim entries belong to the xi-values, while the last dim entries to the eta-values.
    '''
    resultx, resulty = [], []
    for k in range(dim):
        ek = [0 if i != k else 1 for i in range(dim)]
        if not cartesian:
            xi_k = liepoly(a=ek, b=[0]*dim, dim=dim, **kwargs)
            eta_k = liepoly(a=[0]*dim, b=ek, dim=dim, **kwargs)
        else:
            sqrt2 = float(np.sqrt(2))
            xi_k = liepoly(values={tuple(ek + [0]*dim): 1/sqrt2,
                                   tuple([0]*dim + ek): 1/sqrt2},
                                   dim=dim, **kwargs)
            eta_k = liepoly(values={tuple(ek + [0]*dim): -1j/sqrt2,
                                    tuple([0]*dim + ek): 1j/sqrt2},
                                    dim=dim, **kwargs)
        resultx.append(xi_k)
        resulty.append(eta_k)
    return resultx + resulty


def construct(lps, f, power=float('inf')):
    r'''
    Let z = [z1, ..., zk] be Lie polynomials and f an analytical function, taking k values.
    Depending on the input, this routine will either return the Lie polynomial :f(z1, ..., zk): or
    the map f(z1, ..., zk).
    
    Parameters
    ----------
    lps: liepoly or list of liepoly objects
        The Lie polynomial(s) to be constructed.
        
    f: callable
        A function on which we want to apply the list of liepoly objects.
        It needs to be supported by the njet module.
        
    power: int, optional
        The maximal power of the resulting Lie polynomial (default: float('inf')).
        If a value is provided, the routine will return a class of type liepoly, representing
        a Lie polynomial. If nothing is provided, the routine will return the function
        f(z1, ..., zk)
        
    Returns
    -------
    callable or liepoly
        As described above, depending on the 'power' input parameter, either the map f(z1, ..., zk) or
        the Lie polynomial :f(z1, ..., zk): is returned.
    '''
    if lps.__class__.__name__ == 'liepoly':
        lps = [lps]
    n_args_f = len(lps)
    dim_poly = lps[0].dim
    
    assert n_args_f == f.__code__.co_argcount, 'Input function depends on a different number of arguments.'
    assert all([lp.dim == dim_poly for lp in lps]), 'Input polynomials not all having the same dimensions.'

    construction = lambda z: f(*[lps[k](z) for k in range(n_args_f)])   
    if power == float('inf'):
        return construction
    else:
        dcomp = derive(construction, order=power, n_args=2*dim_poly)
        taylor_coeffs = dcomp([0]*2*dim_poly, mult_drv=False)
        return liepoly(values=taylor_coeffs, dim=dim_poly, max_power=power)

class lieoperator:
    '''
    Class to construct and work with an operator of the form g(:x:).
    
    Parameters
    ----------
    x: liepoly
        The function in the argument of the Lie operator.
    
    **kwargs
        Optional arguments may be passed to self.set_generator, self.calcOrbits and self.calcFlow.
    '''
    def __init__(self, x, **kwargs):
        self.init_kwargs = kwargs
        self.flow_parameter = kwargs.get('t', -1) # can be changed in self.calcFlow
        self.set_argument(x, **kwargs)
        
        if 'generator' in kwargs.keys():
            self.set_generator(**kwargs)
        if 'components' in kwargs.keys():
            _ = self.calcOrbits(**kwargs)
        if 't' in kwargs.keys():
            _ = self.calcFlow(**kwargs)
            
    def set_argument(self, x, **kwargs):
        assert x.__class__.__name__ == 'liepoly'
        self.argument = x
        self.n_args = 2*self.argument.dim        
        
    def set_generator(self, generator, **kwargs):
        '''
        Define the generating series for the function g.
        
        Parameters
        ----------
        generator: subscriptable or callable
            If subscriptable, generator[k] =: a_k defines the generating series for the function
            g so that the Lie operator corresponds to g(:x:).
            
            g(z) = sum_k a_k*z**k.
            
            If g is a callable object, then the a_k's are determined from the Taylor coefficients of
            g. Hereby g must depend on only one parameter and it has to be supported by the njet
            module. Furthermore, the additional (integer) parameter 'power' is required to define the 
            maximal order up to which the Taylor coefficients will be determined.
        '''
        if hasattr(generator, '__iter__'):
            # assume that g is in the form of a series, e.g. given by a generator function.
            self.generator = generator
        elif hasattr(generator, '__call__'):
            if not 'power' in kwargs.keys():
                raise RuntimeError("Generation with callable object requires 'power' argument to be set.")
            # assume that g is a function of one variable which needs to be derived n-times at zero.
            assert generator.__code__.co_argcount == 1, 'Function needs to depend on a single variable.'
            dg = derive(generator, order=kwargs['power'])
            taylor_coeffs = dg([0], mult_drv= False)
            self.generator = [taylor_coeffs.get((k,), 0) for k in range(len(taylor_coeffs))]
        else:
            raise NotImplementedError('Input function not recognized.')
        self.power = len(self.generator) - 1
            
    def action(self, y, **kwargs):
        '''
        Apply the Lie operator g(:x:) to a given lie polynomial y to return the elements
        in the series of g(:x:)y.
        
        Parameters
        ----------
        y: liepoly
            The Lie polynomial on which the Lie operator should be applied on.
            
        Returns
        -------
        list
            List containing g[n]*:x:**n y if g = [g[0], g[1], ...] denotes the underlying generator.
            The list goes on up to the maximal power N determined by self.argument.ad routine (see
            documentation there).
        '''
        assert hasattr(self, 'generator'), 'No generator set (check self.set_generator).'
        assert hasattr(self, 'argument'), 'No Lie polynomial in argument set (check self.set_argument)'
        ad_action = self.argument.ad(y, power=self.power)
        assert len(ad_action) > 0
        # N.B. if self.generator[j] = 0, then k_action[j]*self.generator[j] = {}. It will remain in the list below (because 
        # it always holds len(ad_action) > 0 by construction).
        # This is important when calculating the flow later on. In order to check for this consistency, we have added 
        # the above assert statement.
        return [ad_action[j]*self.generator[j] for j in range(len(ad_action))]
        
    def calcOrbits(self, **kwargs):
        '''
        Compute the summands in the series of the Lie operator g(:x:)y, for every requested y,
        by utilizing the routine self.action.
        
        Parameters
        ----------
        components: list, optional
            List of liepoly objects on which the Lie operator g(:x:) should be applied on.
            If nothing specified, then the canonical xi-coordinates are used.
            
        store: bool, optinal
            If true (default), store the current orbits in the field self.orbits.
            
        **kwargs
            Optional arguments passed to lie.create_coords.
            
        Returns
        -------
        list
            A list containing lists of exponents [g[k]*:x:**k (y) for k in (...)], 
            where y is running over the Lie-operator components.
        '''
        if 'components' in kwargs.keys():
            self.components = kwargs['components']
        elif not hasattr(self, 'components'): # then use the xi-polynomials as components.
            self.components = create_coords(dim=self.argument.dim, **kwargs)[:self.argument.dim] 
        orbits = [self.action(y) for y in self.components]
        if kwargs.get('store', True):
            self.orbits = orbits
        return orbits
    
    def flowFunc(self, t, z):
        '''
        Return the flow function phi(t, z) = [g(-t:x:) y](z).
        
        Parameters
        ----------
        t: float
            The requested t-value of the flow.
            
        z: subscriptable
            The point at which to evaluate the flow.
        
        Returns
        -------
        phi: callable
            The flow of the current Lie operator, as described above.
        '''
        return [sum([self.orbits[k][j](z)*(-t)**j for j in range(len(self.orbits[k]))]) for k in range(len(self.orbits))]
     
    def calcFlow(self, **kwargs):
        '''
        Compute the Lie operators [g(t:x:)]y for a given parameter t, for every y in self.components.
        
        Parameters
        ----------
        t: float (or e.g. numpy array), optional
            Parameter in the argument at which the Lie operator should be evaluated.
            
        **kwargs
            Optional arguments passed to self.calcOrbits, if the orbits were not yet computed.
            
        Returns
        -------
        list
            A list containing the flow of every component function of the Lie-operator.
        '''
        t = kwargs.get('t', self.flow_parameter)
        
        if 'orbits' in kwargs.keys():
            orbits = kwargs['orbits']
        elif not hasattr(self, 'orbits'):
            orbits = self.calcOrbits(**kwargs)
        else:
            orbits = self.orbits
        # N.B. We multiply with the parameter t on the right-hand side, because if t is e.g. a numpy array and
        # standing on the left, then numpy would put the liepoly classes into its array, something we do not want. 
        # Instead, we want to put the numpy arrays into our liepoly class.
        flow = [sum([orbits[k][j]*t**j for j in range(len(orbits[k]))]) for k in range(len(orbits))]
        if kwargs.get('store', True):
            self.orbits = orbits
            self.flow = flow
            self.flow_parameter = t
        return flow

    def evaluate(self, z, **kwargs):
        '''
        Evaluate current flow of Lie operator at a specific point.
        
        Parameters
        ----------
        z: subscriptable
            The vector z in the expression (g(:x:)y)(z)
            
        Returns
        -------
        list
            The values (g(:x:)y)(z) for y in self.components.
        '''
        if 't' in kwargs.keys(): # re-evaluate the flow at the requested flow parameter t.
            flow = self.calcFlow(**kwargs)
        else:
            assert hasattr(self, 'flow'), "Flow needs to be calculated first (check self.calcFlow)."
            flow = self.flow
            
        if hasattr(z, 'shape') and hasattr(z, 'reshape') and hasattr(self.flow_parameter, 'shape'):
            # If it happens that both self.flow_parameter and z have a shape (e.g. if both are numpy arrays)
            # then we reshape z to be able to broadcast z and self.flow_parameter into a common array.
            # After the application of self.flow, a reshape on the result is performed in order to
            # shift the two first indices to the last, so that the @ operator can be applied as expected
            # (see PEP 456).
            # In this way it is possible to compute n coordinate points for m flow parameters, while
            # keeping the current self.flow_parameter untouched.
            # TODO: may need to check speed for various reshaping options.
            trailing_ones = [1]*len(self.flow_parameter.shape)
            z = z.reshape(*z.shape, *trailing_ones)
            result = np.array([flow[k](z) for k in range(len(flow))])
            # Now the result has z.shape in its first len(z.shape) indices. We need to bring the first two
            # indices to the rear in order to have an object by which we can apply the conventional matmul operation(s).
            transp_indices = np.roll(np.arange(result.ndim), shift=-2)
            return result.transpose(transp_indices)
        else:
            return [flow[k](z) for k in range(len(flow))]
        
    def apply(self, z, **kwargs):
        '''
        Apply the current Lie operator g(:x:) onto a single liepoly class or a list of liepoly classes.
        This function is basically intended as a shortcut for the successive call of self.calcOrbits and self.calcFlow.
        
        Parameters
        ----------
        z: liepoly or list of liepoly classes
            The Lie polynomial(s) on which to apply the current Lie operator.
            
        Returns
        -------
        self.flow[0] or self.flow
            Depending on the input, either the liepoly g(:x:)y is returned, or a list g(:x:)y for the given
            liepoly elements y.
        '''
        if z.__class__.__name__ == 'liepoly':
            _ = self.calcOrbits(components=[z], **kwargs)
            return self.calcFlow(**kwargs)[0]
        else:
            _ = self.calcOrbits(components=z, **kwargs)
            return self.calcFlow(**kwargs)

    def __call__(self, z, **kwargs):
        '''
        Compute the result of the current Lie operator g(:x:), applied to either 
        1) a specific point
        2) another Lie polynomial
        3) another Lie operator
        
        Parameters
        ----------
        z: subscriptable or liepoly or lieoperator
            
        **kwargs
            Optional arguments passed to self.calcFlow. Note that if an additional parameter t is passed, 
            then the respective results for g(t*:x:) are calculated.
            
        Returns
        -------
        list or liepoly or lieoperator
            1) If z is a list, then the values (g(:x:)y)(z) for the current liepoly elements y in self.components
            are returned (see self.evaluate).
            2) If z is a Lie polynomial, then the orbit of g(:x:)z will be computed and the flow returned as 
               liepoly class (see self.apply).
            3) If z is a Lie operator f(:y:), then the Lie operator h(:z:) = g(:x:) f(:y:) is returned (see self.compose).
        '''
        if z.__class__.__name__ == 'liepoly':
            return self.apply(z, **kwargs)
        elif z.__class__.__name__ == self.__class__.__name__:
            if hasattr(self, 'bch'):
                return self.bch(z, **kwargs)
            else:
                raise NotImplementedError(f"Composition of two objects of type '{self.__class__.__name__}' not supported.")
        else:
            assert hasattr(z, '__getitem__'), 'Input needs to be subscriptable.'
            if z[0].__class__.__name__ == 'liepoly':
                return self.apply(z, **kwargs)
            else:
                return self.evaluate(z, **kwargs)
            
    def copy(self):
        '''
        Create a copy of the current Lie operator
        
        Returns
        -------
        lieoperator
            A copy of the current Lie operator.
        '''
        kwargs = {}
        kwargs.update(self.init_kwargs)
        kwargs['t'] = self.flow_parameter
        if hasattr(self, 'generator'):
            kwargs['generator'] = self.generator
        out = self.__class__(self.argument, **kwargs)
        if hasattr(self, 'orbits'):
            out.orbits = [o.copy() for o in self.orbits]
            out.components = [c.copy() for c in self.components]
        if hasattr(self, 'flow'):
            out.flow = [l.copy() for l in self.flow]
        return out
    
class lexp(lieoperator):  
    '''
    Class to describe Lie operators of the form
      exp(:x:),
    where :x: is a liepoly class.
    
    In contrast to a general Lie operator, we now have the additional possibility to combine several of these operators using the 'combine' routine.
    '''

    def __init__(self, x, power, **kwargs):
        
        self._compose_power_default = 6 # the default power when composing two Lie-operators (used in self.compose)
        kwargs['generator'] = genexp(power)   
        lieoperator.__init__(self, x=x, **kwargs)
        
    def bch(self, other, **kwargs):
        '''
        Compute the composition of the current Lie operator exp(:x:) with another one exp(:y:), 
        to return the Lie operator exp(:z:) given as
           exp(:z:) = exp(:x:) exp(:y:).
           
        Parameters
        ----------
        z: lieoperator
            The Lie operator z = exp(:y:) to be composed with the current Lie operator from the right.
            
        power: int, optional
            The power in the integration variable, to control the degree of accuracy of the result.
            See also lie.combine routine. If nothing specified, self._compose_power_default will be used.
            
        **kwargs
            Additional parameters sent to lie.combine routine.
            
        Returns
        -------
        lieoperator
            The resulting Lie operator of the composition.
        '''
        assert self.__class__.__name__ == other.__class__.__name__
        kwargs['power'] = kwargs.get('power', self._compose_power_default)
        comb, _ = combine(self.argument, other.argument, max_power=self.argument.max_power, **kwargs)
        return self.__class__(sum(comb.values()), power=self.power)
    
    def __matmul__(self, other):
        return self.bch(other)
    
    
    
def combine(*args, power: int, **kwargs):
    '''
    Compute the Lie polynomials of the Magnus expansion, up to a given order.
    
    Parameters
    ----------
    power: int
        The power in s (s: the variable of integration) up to which we consider the Magnus expansion.
        
    *args
        A series of liepoly objects p_j, j = 0, 1, ..., k which to be combined. They may represent 
        the exponential operators exp(:p_j:).
        
    lengths: list, optional
        An optional list of lengths. If nothing specified, the lengths are assumed to be 1.
        
    **kwargs
        Optional keyworded arguments passed to liepoly instantiation and norsett_iserles routine.
        
    Returns
    -------
    dict
        The resulting Lie-polynomials z_j, j \in \{0, 1, ..., r\}, r := power, so that 
        z := z_0 + z_1 + ... z_r satisfies exp((L0 + ... + Lk):z:) = exp(L0:p_0:) exp(L1:p_1:) ... exp(Lk:p_k:),
        accurately up to the requested power r. Hereby it holds: lengths = [L0, L1, ..., Lk].
        Every z_j belongs to Norsett & Iserles approach to the Magnus series.
        
    hard_edge_chain
        The s-dependent Hamiltonian used to construct z.
    '''
    n_operators = len(args)

    # some consistency checks
    assert n_operators > 0
    assert type(power) == int and power >= 0
    dim = args[0].dim
    assert all([op.dim == dim for op in args]), 'The number of variables of the individual Lie-operators are different.'

    lengths = kwargs.get('lengths', [1]*n_operators)
    kwargs['max_power'] = kwargs.get('max_power', min([op.max_power for op in args]))
    
    # The given Lie-polynomials p_1, p_2, ... are representing the chain exp(:p_1:) exp(:p_2:) ... exp(:p_k:) of Lie operators.
    # This means that the last entry p_k is the operator which is executed first. In the hard-edge model, the first entry
    # belongs to the s-dependent Hamiltonian representing the entire chain of the above operators. Therefore, p_k must come first and so
    # the lists have to be reversed:
    args = args[::-1] 
    lengths = lengths[::-1]
    
    # Build the hard-edge Hamiltonian model.
    all_powers = set([k for op in args for k in op.keys()])    
    hamiltonian_values = {k: hard_edge_chain(values=[hard_edge([args[m].get(k, 0)], lengths={1: lengths[m]}) for m in range(n_operators)]) for k in all_powers}
    hamiltonian = liepoly(values=hamiltonian_values, **kwargs)
        
    # Now perform the integration up to the requested power.
    z_series = norsett_iserles(order=power, hamiltonian=hamiltonian, **kwargs)
    out = {}
    for order, trees in z_series.items():
        out_order = 0
        for tpl in trees: # index corresponds to an enumeration of the trees for the specific order
            lp, factor = tpl
            # lp is a liepoly object. Its keys consist of hard_edge_hamiltonians. However we are only interested in their integrals. Therefore:            
            out_order += liepoly(values={k: v._integral*factor for k, v in lp.items()}, **kwargs)
        out[order] = out_order
    return out, hamiltonian



import numpy as np
import mpmath as mp

from njet.ad import standardize_function
from njet.jet import check_zero
from njet import derive

from .lie import liepoly, create_coords
from .lie import lexp as _lexp
from lieops.linalg.nf import normal_form
from lieops.linalg.matrix import matrix_from_dict


def Omega(mu, a, b):
    '''
    Compute the scalar product of mu and a - b.
    
    Parameters
    ----------
    mu: subscriptable
    a: subscriptable
    b: subscriptable
    
    Returns
    -------
    float
        The scalar product (mu, a - b).
    '''
    return sum([mu[k]*(a[k] - b[k]) for k in range(len(mu))])


def homological_eq(mu, Z, **kwargs):
    '''
    Let e[k], k = 1, ..., len(mu) be actions, H0 := sum_k mu[k]*e[k] and Z a
    polynomial of degree n. Then this routine will solve 
    the homological equation 
    {H0, chi} + Z = Q with
    {H0, Q} = 0.

    Attention: No check whether Z is actually homogeneous or real, but if one of
    these properties hold, then also chi and Q will admit such properties.
    
    Parameters
    ----------
    mu: list
        list of floats (tunes).
        
    Z: liepoly
        Polynomial of degree n.
        
    **kwargs
        Arguments passed to liepoly initialization.
        
    Returns
    -------
    chi: liepoly
        Polynomial of degree n with the above property.
        
    Q: liepoly
        Polynomial of degree n with the above property.
    '''
    chi, Q = liepoly(values={}, dim=Z.dim, **kwargs), liepoly(values={}, dim=Z.dim, **kwargs)
    for powers, value in Z.items():
        om = Omega(mu, powers[:Z.dim], powers[Z.dim:])
        if om != 0:
            chi[powers] = 1j/om*value
        else:
            Q[powers] = value
    return chi, Q


def first_order_nf_expansion(H, power: int=2, z=[], warn: bool=True, n_args: int=0, tol: float=1e-14, 
                             code='numpy', **kwargs):
    '''
    Return the Taylor-expansion of a Hamiltonian H in terms of first-order complex normal form coordinates
    around an optional point of interest. For the notation see my thesis.
    
    Parameters
    ----------
    H: callable
        A real-valued function of 2*n parameters (Hamiltonian).
        
    power: int, optional
        The maximal polynomial power of expansion. Must be >= 2.
    
    z: subscriptable, optional
        A point of interest around which we want to expand. If nothing specified,
        then the expansion will take place around zero.
        
    n_args: int, optional
        If H takes a single subscriptable as argument, define the number of arguments with this parameter.
        
    warn: boolean, optional
        Turn on some basic checks:
        a) Warn if the expansion of the Hamiltonian around z contains gradient terms larger than a specific value. 
        b) Verify that the 2nd order terms in the expansion of the Hamiltonian agree with those from the linear theory.
        
    tol: float, optional
        An optional tolerance for checks.
        
    **kwargs
        Arguments passed to linalg.normal_form routine.
        
    Returns
    -------
    dict
        A dictionary of the Taylor coefficients of the Hamiltonian around z, where the first n
        entries denote powers of xi, while the last n entries denote powers of eta.
        
    dict
        The output of linalg.normal_form routine, providing the linear map information at the requested point.
    '''
    assert power >= 2
    Hst, dim = standardize_function(H, n_args=n_args)
    assert dim%2 == 0, 'Dimension must be even; try passing n_args argument.'
    
    # Step 1 (optional): Construct H locally around z (N.B. shifts are symplectic, as they drop out from derivatives.)
    # This step is required, because later on (at point (+)) we want to extract the Taylor coefficients, and
    # this works numerically only if we consider a function around zero.
    if len(z) > 0:
        H = lambda x: Hst([x[k] + z[k] for k in range(len(z))])
    else:
        z = dim*[0]
        H = Hst
    
    # Step 2: Obtain the Hesse-matrix of H.
    # N.B. we need to work with the Hesse-matrix here (and *not* with the Taylor-coefficients), because we want to get
    # a (linear) map K so that the Hesse-matrix of H o K is in CNF (complex normal form). This is guaranteed
    # if the Hesse-matrix of H is transformed to CNF.
    # Note that the Taylor-coefficients of H in 2nd-order are 1/2*Hesse_matrix. This means that at (++) (see below),
    # no factor of two is required.
    dH = derive(H, order=2, n_args=dim)
    z0 = dim*[0]
    Hesse_dict = dH.hess(z0)
    Hesse_matrix = matrix_from_dict(Hesse_dict, symmetry=1, code=code)
    
    # Optional: Raise a warning in case the shifted Hamiltonian still has first-order terms.
    if warn:
        gradient = dH.grad()
        if any([abs(gradient[k]) > tol for k in gradient.keys()]) > 0:
            print (f'Warning: H has a non-zero gradient around the requested point\n{z}\nfor given tolerance {tol}:')
            print ([gradient[k] for k in sorted(gradient.keys())])

    # Step 3: Compute the linear map to first-order complex normal form near z.
    nfdict = normal_form(Hesse_matrix, **kwargs)
    Kinv = nfdict['Kinv'] # Kinv.transpose()@Hesse_matrix@Kinv is in cnf; K(q, p) = (xi, eta)
    
    # Step 4: Obtain the expansion of the Hamiltonian up to the requested power.
    Kmap = lambda zz: [sum([zz[k]*Kinv[j, k] for k in range(len(zz))]) for j in range(len(zz))] # TODO: implement column matrix class. Attention: Kinv[j, k] must stand on right-hand side, otherwise zz[k] may be inserted into a NumPy array!
    HK = lambda zz: H(Kmap(zz))
    dHK = derive(HK, order=power, n_args=dim)
    results = dHK(z0, mult_drv=False) # mult_drv=False ensures that we obtain the Taylor-coefficients of the new Hamiltonian. (+)
    
    if warn:
        # Check if the 2nd order Taylor coefficients of the derived shifted Hamiltonian agree in complex
        # normal form with the values predicted by linear theory.
        HK_hesse_dict = dHK.hess(Df=results)
        HK_hesse_dict = {k: v for k, v in HK_hesse_dict.items() if abs(v) > tol}
        for k in HK_hesse_dict.keys():
            diff = abs(HK_hesse_dict[k] - nfdict['cnf'][k[0], k[1]]) # (++)
            if diff > tol:
                raise RuntimeError(f'CNF entry {k} does not agree with Hamiltonian expansion: diff {diff} > {tol} (tol).')
        
    return results, nfdict


def bnf(H, order: int=1, tol=1e-14, **kwargs):
    '''
    Compute the Birkhoff normal form of a given Hamiltonian up to a specific order.
    
    Attention: Constants and any gradients of H at z will be ignored. If there is 
    a non-zero gradient, a warning is issued by default.
    
    Parameters
    ----------
    H: callable or dict
        Defines the Hamiltonian to be normalized. If H is of type dict, then it must be of the
        form (e.g. for phase space dimension 4): {(i, j, k, l): value}, where the tuple (i, j, k, l)
        denotes the exponents in xi1, xi2, eta1, eta2.
                
    order: int
        The order up to which we build the normal form. Here order = k means that we compute
        k homogeneous Lie-polynomials, where the smallest one will have power k + 2 and the 
        succeeding ones will have increased powers by 1.
        
    max_power: int, optional
        An optional maximal power to be taken into consideration when applying ad-operations between Lie-polynomials.
        
    power: int, optional
        A maximal power by which we want to expand any exponential ad-series while evaluating Lie operators.
        
    tol: float, optional
        Tolerance below which we consider a value as zero. This will be used when examining the second-order
        coefficients of the given Hamiltonian.
        
    **kwargs
        Keyword arguments are passed to .first_order_nf_expansion routine.
        
    Returns
    -------
    dict
        A dictionary with the following keys:
        nfdict: The output of hte first_order_nf_expansion routine.
        H:      Dictionary denoting the used Hamiltonian.
        H0:     Dictionary denoting the second-order coefficients of H.
        mu:     List of the tunes used (coefficients of H0).
        chi:    List of liepoly objects, denoting the Lie-polynomials which map to normal form.
        Hk:     List of liepoly objects, corresponding to the transformed Hamiltonians.
        Zk:     List of liepoly objects, notation see Lem. 1.4.5. in Ref. [1]. 
        Qk:     List of liepoly objects, notation see Lem. 1.4.5. in Ref. [1].
        
    Reference(s):
    [1]: M. Titze: "Space Charge Modeling at the Integer Resonance for the CERN PS and SPS", PhD Thesis (2019). 
    '''
    power = order + 2 # the maximal power of the homogeneous polynomials chi mapping to normal form.
    max_power = kwargs.get('max_power', order + 2) # the maximal power to be taken into consideration when applying ad-operations between Lie-polynomials. Todo: check default & relation to 'power'
    lo_power = kwargs.get('power', order + 2) # The maximal power by which we want to expand exponential series when evaluating Lie operators. Todo: check default.
    
    if type(H) != dict:
        # obtain an expansion of H in terms of complex first-order normal form coordinates
        taylor_coeffs, nfdict = first_order_nf_expansion(H, power=power, **kwargs)
    else:
        taylor_coeffs = H
        nfdict = {}
        
    # get the dimension (by looking at one key in the dict)
    dim2 = len(next(iter(taylor_coeffs)))
    dim = dim2//2
            
    # define mu and H0. For H0 we skip any (small) off-diagonal elements as they must be zero by construction.
    H0 = {}
    mu = []
    for j in range(dim): # add the second-order coefficients (tunes)
        tpl = tuple([0 if k != j and k != j + dim else 1 for k in range(dim2)])
        muj = taylor_coeffs[tpl]
        assert muj.imag < tol
        muj = muj.real
        H0[tpl] = muj
        mu.append(muj)
    H0 = liepoly(values=H0, dim=dim, max_power=max_power)
    
    # For H, we take the values of H0 and add only higher-order terms (so we skip any gradients (and constants). 
    # Note that the skipping of gradients leads to an artificial normal form which may not have anything relation
    # to the original problem. By default, the user will be informed if there is a non-zero gradient 
    # in 'first_order_nf_expansion' routine.
    H = H0.update({k: v for k, v in taylor_coeffs.items() if sum(k) > 2})
    
    # Induction start (k = 2); get P_3 and R_4. Z_2 is set to zero.
    Zk = liepoly(dim=dim, max_power=max_power) # Z_2
    Pk = H.homogeneous_part(3) # P_3
    Hk = H.copy() # H_2 = H
        
    chi_all, Hk_all = [], [H]
    Zk_all, Qk_all = [], []
    for k in range(3, power + 1):
        chi, Q = homological_eq(mu=mu, Z=Pk, max_power=max_power) 
        if len(chi) == 0:
            # in this case the canonical transformation will be the identity and so the algorithm stops.
            break
        Hk = lexp(-chi, power=lo_power)(Hk)
        # Hk = lexp(-chi, power=k + 1)(Hk) # faster but likely inaccurate; need tests
        Pk = Hk.homogeneous_part(k + 1)
        Zk += Q 
        
        chi_all.append(chi)
        Hk_all.append(Hk)
        Zk_all.append(Zk)
        Qk_all.append(Q)

    # assemble output
    out = {}
    out['nfdict'] = nfdict
    out['H'] = H
    out['H0'] = H0
    out['mu'] = mu    
    out['chi'] = chi_all
    out['Hk'] = Hk_all
    out['Zk'] = Zk_all
    out['Qk'] = Qk_all
        
    return out


# We now extend the lexp class with the normal form analysis functionality
class lexp(_lexp):
    
    def __init__(self, *args, **kwargs):
        self.code = kwargs.get('code', 'numpy')
        # The following two internal routines are optional transformations before and after self.__call__ is executed.
        # they are used to conveniently switch the representation of a Lie operator between
        # certain coordinate systems.
        self.transform('default') # to set self._inp and self._out to be used in self.evaluate.
        _lexp.__init__(self, *args, **kwargs)
            
    def set_argument(self, H, **kwargs):
        if not H.__class__.__name__ == 'liepoly':
            assert 'order' in kwargs.keys(), "Lie operator initialized with general callable requires 'order' argument to be set." 
            self.order = kwargs['order']
            # obtain an expansion of H in terms of complex first-order normal form coordinates
            taylor_coeffs, self.nfdict = first_order_nf_expansion(H, code=self.code, **kwargs)
            _lexp.set_argument(self, x=liepoly(values=taylor_coeffs, **kwargs)) # max_power may be set here.
        else: # original behavior
            _lexp.set_argument(self, x=H, **kwargs)
            
    def bnf(self, order, output=True, **kwargs):
        '''
        Compute the Birkhoff normal form of the current Lie exponential operator.
        
        Parameters
        ----------
        order: int
            Order up to which the normal form should be computed.
            
        **kwargs
            Optional arguments passed to 'bnf' routine.
        '''
        return bnf(self.argument._values, order=order, power=self.power, 
                  max_power=self.argument.max_power, **kwargs)
            
    def transform(self, label='', inp=True, out=True, **kwargs):
        '''
        Transform the input and output of self.__call__ into a different coordinate system.
        
        Pre-defined coordinate systems require self.nfdict to be set, given by the
        output of the linalg.normal_form routine.
        
        Parameters
        ----------
        label: str
            The name of the transformation to be used. Currently supported:
            
            1) cnf, default, complex_normal_form
            2) ops, ordinary_phase_space
            3) rnf, real_normal_form, floquet
            4) aa, angle_action
            
            if left blank, then a custom transformation is assumed.
            
        inp: boolean
            Whether the transformation should be applied to the input before self.__call__ (default: True).
            
        out: boolean
            Whether the transformation should be applied to the output after self.__call__ (default: True).
        '''
        if label in ['cnf', 'default', 'complex_normal_form']:
            _inp = lambda z: z
            _out = lambda z: z

        elif label in ['ops', 'ordinary_phase_space']:
            if 'nfdict' in kwargs.keys():
                nfdict = kwargs['nfdict']
            else:
                nfdict = self.nfdict
            _inp = lambda z: nfdict['K']@z # z = (p, q) => U*S*z = (xi, eta) ; K = U*S (see notation in linalg.normal_form)
            _out = lambda z: nfdict['Kinv']@z
        
        elif label in ['rnf', 'real_normal_form', 'floquet']:
            if 'nfdict' in kwargs.keys():
                nfdict = kwargs['nfdict']
            else:
                nfdict = self.nfdict
            _inp = lambda z: nfdict['U']@z # z = (u, v) => U*z = (xi, eta) (see notation in linalg.normal_form)
            _out = lambda z: nfdict['Uinv']@z
        
        elif label in ['aa', 'angle_action']:
            # assume that the first dim parameters are angles, and the last dim parameters are actions.
            dim = kwargs.get('dim', self.n_args//2)
            code = kwargs.get('code', self.code)
            if code == 'mpmath':
                sqrt = mp.sqrt
                exp = mp.exp
                log = mp.log
            else:
                sqrt = np.sqrt
                exp = np.exp
                log = np.log
                
            def _inp(z):
                angle, action = z[:dim], z[dim:]
                xi, eta = [], []
                for k in range(dim):
                    xi.append(sqrt(action[k])*exp(-1j*angle[k]))
                    eta.append(sqrt(action[k])*exp(1j*angle[k]))
                return xi + eta
            
            def _out(z):
                xi, eta = z[:dim], z[dim:]
                angle, action = [], []
                for k in range(dim):          
                    action.append(xi[k]*eta[k])
                    if not check_zero(xi[k]):
                        angle.append(-1j*log(eta[k]/xi[k])/2)
                    elif not check_zero(eta[k]): 
                        angle.append(1j*log(xi[k]/eta[k])/2) 
                    else:
                        angle.append(0) # default value for 0-actions
                return angle + action
            
        else:
            # User-defined transformation
            inp, out = False, False
            assert 'T' in kwargs.keys() or 'Tinv' in kwargs.keys(), f"Custom transformation requires 'T' or 'Tinv' parameters to be set."
            if 'T' in kwargs.keys():
                assert hasattr(T, '__call__'), 'T not callable.'
                inp = True
                _inp = T
            if 'Tinv' in kwargs.keys():
                assert hasattr(Tinv, '__call__'), 'Tinv not callable.'
                out = True
                _out = Tinv
            
        if inp:
            self._inp = _inp
        else:
            self._inp = lambda z: z
            
        if out:
            self._out = _out
        else:
            self._out = lambda z: z
                
    def evaluate(self, z, **kwargs):
        z = self._inp(z)
        return self._out(_lexp.evaluate(self, z=z, **kwargs))

        


# This script collects (or loads) routines which are more fundamental and are
# often required for more sophisticated routines.

import numpy as np
import mpmath as mp
from scipy.linalg import schur

from .matrix import get_package_name

def twonorm(vector, mode='complex', **kwargs):
    # Compute the 2-norm of a vector.
    # This seems to provide slightly faster results than np.linalg.norm
    if mode == 'complex':
        sum2 = sum([vector[k].conjugate()*vector[k] for k in range(len(vector))])
    else:
        sum2 = sum([vector[k]*vector[k] for k in range(len(vector))])
        
    code = get_package_name(sum2)
    
    if code == 'mpmath':
        return mp.sqrt(mp.re(sum2))
    if code in ['numpy', 'builtins']:
        return np.sqrt(np.real(sum2))
    
def gram_schmidt(vectors, mode='complex', tol=1e-15, **kwargs):
    '''Gram-Schmidt orthogonalization procedure of linarly independent vectors with complex entries, i.e.
    'unitarization'.
    
    Parameters
    ----------
    vectors: list
        list of vectors to be orthogonalized.
        
    mode: str, optional
        If mode == 'complex' (default), then all norms and scalar products are computed using conjugation.
        
    **kwargs 
        Additional arguments passed to .twonorm
    
    Returns
    -------
    list
        list of length len(vectors) which are mutually unitary. I.e.
        with O := np.array(list).T it holds O^H@O = 1, where ^H means transposition and complex conjugation.
    '''
    k = len(vectors)
    dim = len(vectors[0])
    norm_0 = twonorm(vectors[0], mode=mode, **kwargs)
    ortho = {(m, 0): vectors[0][m]/norm_0 for m in range(dim)} # ortho[(m, j)] denotes the m-th component of the j-th (new) vector
    for i in range(1, k):
        sum1 = {(m, i): vectors[i][m] for m in range(dim)}
        for j in range(i):
            if mode == 'complex':
                scalar_product_ij = sum([ortho[(r, j)].conjugate()*vectors[i][r] for r in range(dim)])
            else:
                scalar_product_ij = sum([ortho[(r, j)]*vectors[i][r] for r in range(dim)])
                
            for m in range(dim):
                sum1[(m, i)] -= scalar_product_ij*ortho[(m, j)]  
        norm_i = twonorm([sum1[(m, i)] for m in range(dim)], mode=mode, **kwargs)
        if abs(norm_i) < tol:
            raise RuntimeError(f'Division by zero with mode ({mode}) encountered; check input on linearly independence.')
        for m in range(dim):
            ortho[(m, i)] = sum1[(m, i)]/norm_i
    return [[ortho[(m, i)] for m in range(dim)] for i in range(k)]


def rref(M, augment=None, tol=1e-10, **kwargs):
    '''
    Compute the reduced row echelon form of M (M can be a real or complex matrix).
    
    Parameters
    ----------
    M: matrix
        matrix to be transformed.
        
    augment: matrix, optional
        optional matrix to be used on the right-hand side of M, and which will be simultaneously transformed.
        If nothing specified, the identity matrix will be used.
        
    tol: float, optional
        A tolerance by which we identify small numbers as zero.
    
    Returns
    -------
    triple
        A triple consisting of i) the transformed matrix M, ii) the transformed augment and iii) the pivot indices
        of the transformed matrix M (a list of tuples).
    '''
    code = get_package_name(M)
    # reduced row echelon form of M
    if code == 'numpy':
        n, m = M.shape
        if augment == None:
            augment = np.eye(n)        
        assert augment.shape[0] == n
        Mone = np.bmat([M, augment])
        
    if code == 'mpmath':
        n, m = M.rows, M.cols
        if augment == None:
            augment = mp.eye(n)
        assert augment.rows == n
        Mone = mp.zeros(n, m + augment.cols)
        Mone[:, :m] = M
        Mone[:, m:] = augment
    
    # transform Mone = (M | 1)
    pivot_row_index = 0
    pivot_indices = [] # to record the pivot indices
    for j in range(m):
        
        # Step 1: determine the pivot row of column j
        column_j_has_pivot = False
        for k in range(pivot_row_index, n):
            # skip 0-entries
            if abs(Mone[k, j]) < tol:
                continue
                
            if k > pivot_row_index:
                # exchange this entry with those at the intended pivot_row_index row
                if code == 'numpy':
                    pivot_row = np.copy(Mone[pivot_row_index, :])
                    Mone[pivot_row_index, :] = np.copy(Mone[k, :])/Mone[k, j]
                elif code == 'mpmath':
                    pivot_row = Mone[pivot_row_index, :].copy()
                    Mone[pivot_row_index, :] = Mone[k, :].copy()/Mone[k, j]
                Mone[k, :] = pivot_row

            # normalize the column belonging to the pivot index
            Mone[pivot_row_index, :] = Mone[pivot_row_index, :]/Mone[pivot_row_index, j]
                    
            pivot_indices.append((pivot_row_index, j)) # record the pivot indices for output
            column_j_has_pivot = True
            break
                
        if not column_j_has_pivot: 
            # It can happen that column_j_has_pivot = False: Namely, if the colum has only zero entries.
            # In this case we just proceed with the next column.
            continue
            
        # Step 2: eliminate the other non-zero rows.
        for k in range(0, n):
            if abs(Mone[k, j]) < tol or k == pivot_row_index:
                continue
            Mone[k, :] = Mone[k, :] - Mone[k, j]*Mone[pivot_row_index, :]
            
        pivot_row_index = pivot_row_index + 1
        
    return Mone[:,:m], Mone[:,m:], pivot_indices


def imker(M, **kwargs):
    '''
    Obtain a basis for the image and the kernel of M.
    M can be a real or complex matrix.
    
    Parameters
    ----------
    M:
        matrix to be analyzed.
        
    **kwargs
        Additional arguments passed to rref routine.
        
    Returns
    -------
    image:
        matrix spanning the image of M
        
    kernel:
        matrix spanning to the kernel of M
    '''
    # Idea taken from
    # https://math.stackexchange.com/questions/1612616/how-to-find-null-space-basis-directly-by-matrix-calculation    

    code = get_package_name(M)
    if code == 'numpy':
        M = np.array(M)
    if code == 'mpmath':
        M = mp.matrix(M)
    
    ImT, KerT, pivots = rref(M.transpose(), **kwargs) # transpose the input matrix to obtain kernel & image in the end.
    if len(pivots) == 0:
        # this can happen if M = 0, so no pivot points exist.
        if code == 'numpy':
            kernel = np.eye(M.shape[1])
            image = np.matrix([[]])
        elif code == 'mpmath':
            kernel = mp.eye(M.cols)
            image = mp.matrix([[]])
    else: 
        zero_row_indices = pivots[-1][0] + 1
        kernel = KerT[zero_row_indices:, :].transpose()
        image = ImT[:zero_row_indices, :].transpose()
        
    # consistency check
    if code == 'numpy':
        assert kernel.shape[1] + image.shape[1] == M.shape[1]
    if code == 'mpmath':
        assert kernel.cols + image.cols == M.cols        
        
    return image, kernel


def basis_extension(*vects, gs=False, **kwargs):
    '''
    Provide an extension of a given set of vectors to span the full space. 
    The vectors can have real or complex coefficients.
    
    Parameters
    ----------
    *vects:
        Vectors to be extended
        
    gs: boolean, optional
        Apply the Gram-Schmidt orthogonalization procedure on the extension.
        
    **kwargs
        Optional arguments passed to gram_schmidt. In use only if gs=True.
        
    Returns
    -------
    matrix
        A matrix representing a basis extension of the given vectors.
    '''
    code = get_package_name(vects[0])
    if code == 'numpy':
        vects = np.array(vects)
    if code == 'mpmath':
        vects = mp.matrix(vects)
        
    _, ext = imker(vects.conjugate(), **kwargs)
    
    if code == 'numpy':
        n, m = ext.shape
    elif code == 'mpmath':
        n, m = ext.rows, ext.cols
        
    if gs and n > 0 and m > 0:
        ext = gram_schmidt([[ext[k, j] for k in range(n)] for j in range(m)], **kwargs)
        
    if code == 'numpy':
        return np.array(ext).transpose()
    elif code == 'mpmath':
        return mp.matrix(ext).transpose()


def eig(M, symmetric=False, **kwargs):
    '''
    Compute the eigenvalues and eigenvectors of a given matrix, based on underlying code.
    
    Parameters
    ----------
    M: matrix
        Matrix to be considered.
        
    symmetric: boolean, optional
        Whether the matrix is assumed to be symmetric.
    '''
    code = get_package_name(M)
    
    if code == 'numpy':
        assert M.shape[0] == M.shape[1]
        if not symmetric:
            eigenvalues, eigenvectors = np.linalg.eig(M)
        else:
            eigenvalues, eigenvectors = np.linalg.eigh(M)
        eigenvalues = eigenvalues.tolist()
        eigenvectors = [np.array(eigenvectors)[:, j] for j in range(len(eigenvalues))] # convert the np.matrix objects, which are the result of the
        # numpy eig routine back to np.array's.
        
    if code == 'mpmath':
        assert M.cols == M.rows
        if 'dps' in kwargs.keys():
            mp.mp.dps = kwargs['dps'] # number of digits defining precision.#

        if not symmetric:
            eigenvalues, eigenvectors = mp.eig(mp.matrix(M))    
        else:
            eigenvalues, eigenvectors = mp.eigh(mp.matrix(M))
        eigenvectors = [eigenvectors[:, j] for j in range(len(eigenvalues))]
        
    return eigenvalues, eigenvectors


def eigenspaces(M, flatten=False, tol=1e-10, check=True, **kwargs):
    '''
    Let M be a square matrix. Then this routine will determine a basis of normalized eigenvectors. Hereby
    eigenvectors belonging to the same eigenvalues are (complex) orthogonalized.
    
    Parameters
    ----------
    M:
        list of vectors defining a (n x n)-matrix.
        
    flatten: boolean, optional
        If True, flatten the respective results (default: False).
        
    tol: float, optional
        Parameter to identify small values as being zero.
        
    check: boolean, optional
        Check if the number of zero-eigenvalues is consistent with the dimension of the kernel of the input matrix within the given tolerance.
        The kernel of the input matrix is hereby determined by the imker routine.
        
    **kwargs
        Optional arguments passed to eig, imker and gram_schmidt routines.
        
    Returns
    -------
    eigenvalues: list
        List of elements, where the k-th element constitute the eigenvalue to the k-th eigenspace.
    
    eigenvectors: list
        List of lists, where the k-th element is a list of pairwise unitary vectors spanning the k-th eigenspace.
    '''
    code = get_package_name(M)
    eigenvalues, eigenvectors = eig(M, **kwargs)
        
    # consistency check
    n = len(eigenvalues)
    assert n > 0
    dim = len(eigenvectors[0])
    assert all([twonorm(e) >= tol for e in eigenvectors]), f'Norm of at least one eigenvector < {tol}.' # sometimes mpmath produces eigenvectors of zero norm!
        
    # group the indices of the eigenvectors if they belong to the same eigenvalues.
    eigenspaces = [[0]] # 'eigenspaces' will be the collection of these groups of indices.
    for i in range(1, n):
        j = 0
        while j < len(eigenspaces):
            if abs(eigenvalues[i] - eigenvalues[eigenspaces[j][0]]) < tol: 
                # eigenvalues[i] has been identified to belong to eigenspaces[j] group; append the index to this group
                eigenspaces[j].append(i)
                break
            j += 1    
        if j == len(eigenspaces): 
            # no previous eigenspace belonging to this eigenvalue has been found; create a new group
            eigenspaces.append([i])
    
    if check:
        # check if we have identified the number of zero-eigenvalues 
        # to agree with the dimension of the kernel of the input matrix
        image, kernel = imker(M, tol=tol, **kwargs)
        if code == 'numpy':
            dim_kernel = kernel.shape[1]
        if code == 'mpmath':
            dim_kernel = kernel.cols
        # check if tolerance can detect the zero-eigenvalues
        n_zero_eigenvalues = len([e for e in eigenspaces if abs(eigenvalues[e[0]]) < tol])
        assert dim_kernel == n_zero_eigenvalues, f'The number of zero-eigenvalues ({n_zero_eigenvalues}) is not consistent with the dimension ({dim_kernel}) of the kernel of the input matrix, both determined using a tolerance of {tol}. Check input and/or try to adjust precision.'
                
    # orthogonalize vectors within the individual eigenspaces
    eigenvalues_result, eigenvectors_result = [], []
    for indices in eigenspaces:
        vectors = [[eigenvectors[k][j] for k in indices] for j in range(dim)]
        # the vectors given by the eig routine may be linearly dependent; we therefore need to orthogonalize its image
        if code == 'numpy':
            span = np.array(vectors)
        if code == 'mpmath':
            span = mp.matrix(vectors)
        vimage, vkernel = imker(span, tol=tol, **kwargs)
        
        # creates a list of *column*-vectors, as required by gram_schmidt routine. TODO: need way to treat both codes simultaneously...
        if code == 'numpy':
            vimage = vimage.transpose().tolist()
        if code == 'mpmath':
            vimage = [[vimage[j, k] for j in range(vimage.rows)] for k in range(vimage.cols)]

        on_vectors = gram_schmidt(vimage, tol=tol, **kwargs)
        on_eigenvalues = [eigenvalues[k] for k in indices[:len(on_vectors)]]
        if flatten:
            eigenvectors_result += on_vectors
            eigenvalues_result += on_eigenvalues
        else:
            eigenvectors_result.append(on_vectors)
            eigenvalues_result.append(on_eigenvalues[0]) # all of these eigenvalues are considered to be equal, so we pic the first one.
            
    return eigenvalues_result, eigenvectors_result


def get_principal_sqrt(M, **kwargs):
    '''
    A numeric algorithm to obtain the principal square root of a matrix M (if it exists).
    The principal square is the unique square root of a matrix so that -pi/2 < arg(l(X)) <= pi/2
    holds, where
    It can can be expressed in terms of a polynomial in M.
    Refernce: Ake Björk: Numerical Methods in Matrix Computations (2014), sec. 3.8.1.
    '''
    T, Q = schur(M, **kwargs) # Q@T@Q.transpose().conjugate() = M 
    # N.B. Q is a unitary matrix
    
    dim = len(T)
    S = np.diag(np.sqrt(T.diagonal(), dtype=complex))
    for diagonal in range(1, dim): # diagonal = 1: first off-diagonal S[0, 1], S[1, 2], ... etc.
        for diagonal_index in range(dim - diagonal):
            i = diagonal_index
            j = diagonal + diagonal_index        
            S[i, j] = (T[i, j] - sum([S[i, k]*S[k, j] for k in range(i + 1, j)]))/(S[i, i] + S[j, j]) # the j - 1 in the Eq. in the reference must be increased by 1 here due to Python convention.

    # check: S@S = T
    return Q@S@Q.transpose().conjugate()



import numpy as np

from njet.jet import factorials, check_zero
from njet import derive

from .genfunc import genexp

class liepoly:
    '''
    Class to model the Lie operator :p:, where p is a polynomial given in terms of
    complex (xi, eta)-coordinates. For the notation of these coordinates see Ref.
    "Malte Titze: Space Charge Modeling at the Integer Resonance for the CERN PS and SPS" (2019),
    p. 33 onwards.
    
    Parameters
    ----------
    self.max_power: int
        A value > 0 means that any calculations leading to expressions beyond this 
        degree will be discarded. For binary operations the minimum of both 
        max_powers are used.
        
    **kwargs
        Optional arguments passed to self.set_monimial
    '''
    def __init__(self, max_power=float('inf'), **kwargs):
        # self.dim denotes the number of xi (or eta)-factors.
        if 'values' in kwargs.keys():
            self.values = kwargs['values']
        elif 'a' in kwargs.keys() or 'b' in kwargs.keys():
            self.set_monomial(**kwargs)
        else:
            self.values = {}
            
        if len(self.values) == 0:
            self.dim = kwargs.get('dim', 0)
        else:
            self.dim = kwargs.get('dim', len(next(iter(self.values)))//2)
            
        self.set_max_power(max_power)
        
    def set_max_power(self, max_power):
        self.max_power = max_power
        self.values = {k: v for k, v in self.values.items() if sum(k) <= max_power}
        
    def set_monomial(self, a=[], b=[], value=1, **kwargs):
        dim = max([len(a), len(b)])
        if len(a) < dim:
            a += [0]*(dim - len(a))
        if len(b) < dim:
            b += [0]*(dim - len(b))
        self.values = {tuple(a + b): value}
        
    def maxdeg(self):
        '''
        Obtain the maximal degree of the current Lie polynomial. 
        '''
        if len(self.values) == 0:
            return 0
        else:
            return max([sum(k) for k, v in self.values.items()])
    
    def mindeg(self):
        '''
        Obtain the minimal degree of the current Lie polynomial. 
        '''
        if len(self.values) == 0:
            return 0
        else:
            return min([sum(k) for k, v in self.values.items()])
        
    def copy(self):
        return self.__class__(values={k: v for k, v in self.values.items()}, dim=self.dim, max_power=self.max_power)
    
    def extract(self, condition):
        '''
        Extract a Lie polynomial from the current Lie polynomial, based on a condition.
        
        Parameters
        ----------
        condition: callable
            A function which maps a given tuple (an index) to a boolean. For example 'condition = lambda x: sum(x) == k' would
            yield the homogeneous part of the current Lie polynomial (this is realized in 'self.homogeneous_part').
            
        Returns
        -------
        liepoly
            The extracted Lie polynomial.
        '''
        return self.__class__(values={key: value for key, value in self.values.items() if condition(key)}, dim=self.dim, max_power=self.max_power)
    
    def homogeneous_part(self, k: int):
        '''
        Extract the homogeneous part of order k from the current Lie polynomial.
        
        Parameters
        ----------
        k: int
            The requested order.
            
        Returns
        -------
        liepoly
            The extracted Lie polynomial.
        '''
        return self.extract(condition=lambda x: sum(x) == k)
        
    def __call__(self, z):
        '''
        Evaluate the polynomial at a specific position z.
        
        Parameters
        ----------
        z: subscriptable
            The point at which the polynomial should be evaluated. It is assumed that len(z) == self.dim
            (the components of z correspond to the xi-values).
        '''
        assert len(z) == 2*self.dim, f'Number of input parameters: {len(z)}, expected: {2*self.dim}'
        result = 0
        for k, v in self.values.items():
            prod = 1
            for j in range(self.dim):
                prod *= z[j]**k[j]*z[j + self.dim]**k[j + self.dim]
            result += prod*v # v needs to stay on the right-hand side here, because prod may be a jet class (if we
            # compute the derivative(s) of the Lie polynomial)
        return result
        
    def __add__(self, other):
        add_values = {k: v for k, v in self.values.items()}
        if self.__class__.__name__ != other.__class__.__name__:
            # Treat other object as constant.
            if other != 0:
                zero_tpl = (0,)*self.dim*2
                new_value = add_values.get(zero_tpl, 0) + other
                if not check_zero(new_value):
                    add_values[zero_tpl] = new_value
                else:
                    _ = add_values.pop(zero_tpl, None)
            max_power = self.max_power
        else:
            assert other.dim == self.dim
            for k, v in other.values.items():
                new_v = add_values.get(k, 0) + v
                if not check_zero(new_v):
                    add_values[k] = new_v
                else:
                    _ = add_values.pop(k, None)
            max_power = min([self.max_power, other.max_power])
        return self.__class__(values=add_values, dim=self.dim, max_power=max_power)
    
    def __radd__(self, other):
        return self + other
    
    def __neg__(self):
        return self.__class__(values={k: -v for k, v in self.values.items()}, 
                              dim=self.dim, max_power=self.max_power)
    
    def __sub__(self, other):
        return self + -other

    def __matmul__(self, other):
        return self.poisson(other)
        
    def poisson(self, other):
        '''
        Compute the Poisson-bracket {self, other}
        '''
        assert self.__class__.__name__ == other.__class__.__name__
        assert other.dim == self.dim
        max_power = min([self.max_power, other.max_power])
        poisson_values = {}
        for t1, v1 in self.values.items():
            power1 = sum(t1)
            for t2, v2 in other.values.items():
                power2 = sum(t2)
                if power1 + power2 - 2 > max_power:
                    continue
                a, b = t1[:self.dim], t1[self.dim:]
                c, d = t2[:self.dim], t2[self.dim:]
                for k in range(self.dim):
                    det = a[k]*d[k] - b[k]*c[k]
                    if det == 0:
                        continue
                    new_power = tuple([a[j] + c[j] if j != k else a[j] + c[j] - 1 for j in range(self.dim)] + \
                                [b[j] + d[j] if j != k else b[j] + d[j] - 1 for j in range(self.dim)])
                    new_value = poisson_values.get(new_power, 0) - 1j*det*v1*v2
                    if not check_zero(new_value):
                        poisson_values[new_power] = new_value
                    else:
                        _ = poisson_values.pop(new_power, None)
        return self.__class__(values=poisson_values, dim=self.dim, max_power=max_power)
    
    def __mul__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            assert self.dim == other.dim
            dim2 = 2*self.dim
            max_power = min([self.max_power, other.max_power])
            mult_values = {}
            for t1, v1 in self.values.items():
                power1 = sum(t1)
                for t2, v2 in other.values.items():
                    power2 = sum(t2)
                    if power1 + power2 > max_power:
                        continue
                    prod_tpl = tuple([t1[k] + t2[k] for k in range(dim2)])
                    prod_val = mult_values.get(prod_tpl, 0) + v1*v2 # it is assumed that v1 and v2 are both not zero, hence prod_val != 0.
                    if not check_zero(prod_val):
                        mult_values[prod_tpl] = prod_val
                    else:
                        _ = mult_values.pop(prod_tpl, None)
            return self.__class__(values=mult_values, dim=self.dim, max_power=max_power)            
        else:
            return self.__class__(values={k: v*other for k, v in self.values.items() if not check_zero(other)}, 
                                          dim=self.dim, max_power=self.max_power)
        
    def __rmul__(self, other):
        return self.__mul__(other)
    
    def __truediv__(self, other):
        # implement '/' operator
        if self.__class__.__name__ != other.__class__.__name__:
            # Attention: If other is a NumPy array, there is no check if one of the entries is zero.
            return self.__class__(values={k: v/other for k, v in self.values.items() if not check_zero(other)}, 
                                          dim=self.dim, max_power=self.max_power)
        else:
            raise NotImplementedError('Division by Lie polynomial not supported.')
        
    
    def __pow__(self, other):
        assert type(other) == int 
        assert other >= 0
        if other == 0:
            return self.__class__(values={(0,)*self.dim*2: 1}, 
                                  dim=self.dim, max_power=self.max_power) # N.B. 0**0 := 1

        remainder = other%2
        half = self**(other//2)
        if remainder == 1:
            return self*half*half
        else:
            return half*half
        
    def __len__(self):
        return len(self.values)
    
    def __eq__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            return self.values == other.values
        else:
            if self.maxdeg() != 0:
                return False
            else:
                return self.values.get((0, 0), 0) == other
            
    def __getitem__(self, key):
        return self.values[key]
        
    def ad(self, y, power: int=1):
        '''
        Compute repeated Poisson-brackets.
        E.g. let x = self. Then {x, {x, {x, {x, {x, {x, y}}}}}} =: x**6(y)
        Special case: x**0(y) := y
        
        Let z be a homogeneous Lie polynomial and deg(z) the degree of z. Then it holds
        deg(x**m(y)) = deg(y) + m*(deg(x) - 2).
        This also holds for the maximal degrees and minimal degrees in case that x, and y
        are inhomogeneous.
        
        Therefore, if x and y both have non-zero max_power fields, 'self.power' will not evaluate
        terms x**m with
        m >= min([(max_power - mindeg(y))//(mindeg(x) - 2), power]).
        
        Parameters
        ----------
        y: liepoly
            Lie polynomial which we want to evaluate on

        power: int, optional
            Number of repeated brackets (default: 1).


        Returns
        -------
        list
            List [x**k(y) for k in range(n + 1)], if n is the power requested.
        '''
        assert self.__class__.__name__ == y.__class__.__name__
        assert power >= 0
        
        # Adjust requested power if max_power makes this necessary, see comment above.
        max_power = min([self.max_power, y.max_power])
        mindeg_x = self.mindeg()
        if mindeg_x > 2 and max_power < float('inf'):
            mindeg_y = y.mindeg()
            power = min([(max_power - mindeg_y)//(mindeg_x - 2), power]) # N.B. // works as floor division
            
        result = self.__class__(values={k: v for k, v in y.values.items()}, 
                                dim=y.dim, max_power=max_power)
        all_results = [result]
        # N.B.: We can not set values = self.values, otherwise result.values will get changed if self.values is changing.
        for k in range(power):
            result = self@result
            if len(result) == 0:
                break
            all_results.append(result)
        return all_results
    
    def __str__(self):
        return self.values.__str__()

    def _repr_html_(self):
        return f'<samp>{self.__str__()}</samp>'
    
    def derive(self, **kwargs):
        '''
        Derive the current Lie polynomial.
        
        Parameters
        ----------
        order: int
            The order by which we are going to derive the polynomial.
            
        Returns
        -------
        derive
            A class of type njet.ad.derive with n_args=2*self.dim parameters.
            Note that a function evaluation should be consistent with the fact that 
            the last self.dim entries are the complex conjugate values of the 
            first self.dim entries.
        '''
        return derive(self, n_args=2*self.dim, **kwargs)
    
    def flow(self, power: int, t=1, **kwargs):
        '''
        Let f: R^n -> R be a differentiable function and :x: the current polynomial Lie map. 
        Then this routine will compute the components of M: R^n -> R^n,
        where M is the map satisfying
        exp(:x:) f = f o M

        Note that the degree to which powers are discarded is given by self.max_power.

        Parameters
        ----------
        power: int
            The maximal power up to which exp(:x:) should be evaluated.
        
        t: float, optional
            An additional parameter to model exp(t*:x:) (default: 1). Note that
            this parameter can also be changed in the lieoperator class later.
            
        **kwargs
            Additional arguments are passed to the lieoperator class.

        Returns
        -------
        lieoperator
            Class of type lieoperator, modeling the flow of the current Lie polynomial.
        '''
        return lieoperator(self, generator=genexp(power), t=t, **kwargs)
        
    def construct(self, f, **kwargs):
        '''
        Let :x: represent the current Lie polynomial. Depending on the input,
        this routine will either return the map f(x) or the Lie polynomial :f(x):.
        
        Parameters
        ----------
        f: callable
            A function depending on a single parameter. It needs to be supported by the njet module.
            
        **kwargs
            Additional parameters passed to lie.construct routine.
            
        Returns
        -------
        callable or liepoly
            The output depends on the optional argument 'power'.
            
            If no argument 'power' has been passed, then it will
            be taken from the current value self.max_power.
            
            If power < float('inf'), then the Lie polynomial :f(x): is returned,
            where f has been expanded up to the specified power. If power == float('inf'),
            then the function f(x) is returned.
        '''
        if not 'power' in kwargs.keys():
            kwargs['power'] = self.max_power
        return construct([self], f, **kwargs)
            
    
def create_coords(dim, **kwargs):
    '''
    Create a set of complex (xi, eta)-Lie polynomials for a given dimension.
    
    Parameters
    ----------
    dim: int
        The requested dimension.
        
    **kwargs
        Optional arguments passed to liepoly class.
        
    Returns
    -------
    list
        List of length 2*dim with liepoly entries, corresponding to the xi_k and eta_k Lie polynomials. Hereby the first
        dim entries belong to the xi-values, while the last dim entries to the eta-values.
    '''
    resultx, resulty = [], []
    for k in range(dim):
        ek = [0 if i != k else 1 for i in range(dim)]
        xi_k = liepoly(a=ek, b=[0]*dim, dim=dim, **kwargs)
        eta_k = liepoly(a=[0]*dim, b=ek, dim=dim, **kwargs)
        resultx.append(xi_k)
        resulty.append(eta_k)
    return resultx + resulty


def construct(lps, f, power=float('inf')):
    r'''
    Let z = [z1, ..., zk] be Lie polynomials and f an analytical function, taking k values.
    Depending on the input, this routine will either return the Lie polynomial :f(z1, ..., zk): or
    the map f(z1, ..., zk).
    
    Parameters
    ----------
    lps: liepoly or list of liepoly objects
        The Lie polynomial(s) to be constructed.
        
    f: callable
        A function on which we want to apply the list of liepoly objects.
        It needs to be supported by the njet module.
        
    power: int, optional
        The maximal power of the resulting Lie polynomial (default: float('inf')).
        If a value is provided, the routine will return a class of type liepoly, representing
        a Lie polynomial. If nothing is provided, the routine will return the function
        f(z1, ..., zk)
        
    Returns
    -------
    callable or liepoly
        As described above, depending on the 'power' input parameter, either the map f(z1, ..., zk) or
        the Lie polynomial :f(z1, ..., zk): is returned.
    '''
    if lps.__class__.__name__ == 'liepoly':
        lps = [lps]
    n_args_f = len(lps)
    dim_poly = lps[0].dim
    
    assert n_args_f == f.__code__.co_argcount, 'Input function depends on a different number of arguments.'
    assert all([lp.dim == dim_poly for lp in lps]), 'Input polynomials not all having the same dimensions.'

    construction = lambda z: f(*[lps[k](z) for k in range(n_args_f)])   
    if power == float('inf'):
        return construction
    else:
        dcomp = derive(construction, order=power, n_args=2*dim_poly)
        taylor_coeffs = dcomp([0]*2*dim_poly, mult=False)
        return liepoly(values=taylor_coeffs, dim=dim_poly, max_power=power)
    

def exp_ad(x, y, power, **kwargs):
    '''
    Compute the exponential Lie operator exp(:x:)y up to a given power.

    Parameters
    ----------
    x: liepoly
        The Lie polynomial in the exponent of exp.
    
    y: liepoly or list of liepoly objects
        The Lie polynomial(s) on which we want to apply the exponential Lie operator.
        
    power: int
        Integer defining the maximal power up to which we want to compute the result.
        
    **kwargs
        Additional arguments passed to the lieoperator class.

    Returns
    ------- 
    liepoly or list of liepoly objects
        Class of type liepoly, representing the result of the exponential Lie operator exp(:x:)y,
        up to the requested power.
    '''
    if y.__class__.__name__ == 'liepoly': # if the input is a Lie polynomial, also return a Lie polynomial
        return lieoperator(x, generator=genexp(power), components=[y], **kwargs).flow[0]
    else:
        return lieoperator(x, generator=genexp(power), components=y, **kwargs).flow


class lieoperator:
    '''
    Class to construct and work with a Lie operator of the form g(:x:).
    
    Parameters
    ----------
    x: liepoly
        The function in the exponent.
    
    **kwargs
        Optional arguments may be passed to self.set_generator, self.calcOrbits and self.calcFlow.
    '''
    def __init__(self, x, **kwargs):
        self.flow_parameter = 1 # default flow parameter to be used (and may be changed) in self.calcFlow
        self.set_exponent(x, **kwargs)
        if not 'generator' in kwargs.keys() and 'power' in kwargs.keys(): 
            # if only a power argument is given, and no generator specified, 
            # the default Lie operator will be the exponential Lie operator.
            kwargs['generator'] = genexp(kwargs['power'])    
        if 'generator' in kwargs.keys():
            self.set_generator(**kwargs)
        if 'components' in kwargs.keys() or 't' in kwargs.keys():
            self.calcOrbits(**kwargs)
            self.calcFlow(**kwargs)
            
    def set_exponent(self, x, **kwargs):
        assert x.__class__.__name__ == 'liepoly'
        self.exponent = x
        self.n_args = 2*self.exponent.dim        
        
    def set_generator(self, generator, **kwargs):
        '''
        Define the generating series for the function g.
        
        Parameters
        ----------
        generator: subscriptable or callable
            If subscriptable, generator[k] =: a_k defines the generating series for the function
            g so that the Lie operator corresponds to g(:x:).
            
            g(z) = sum_k a_k*z**k.
            
            If g is a callable object, then the a_k's are determined from the Taylor coefficients of
            g. Hereby g must depend on only one parameter and it has to be supported by the njet
            module. Furthermore, the additional (integer) parameter 'power' is required to define the 
            maximal order up to which the Taylor coefficients will be determined.
        '''
        if hasattr(generator, '__iter__'):
            # assume that g is in the form of a series, e.g. given by a generator function.
            self.generator = generator
        elif hasattr(generator, '__call__'):
            if not 'power' in kwargs.keys():
                raise RuntimeError("Generation with callable object requires 'power' argument to be set.")
            # assume that g is a function of one variable which needs to be derived n-times at zero.
            assert generator.__code__.co_argcount == 1, 'Function needs to depend on a single variable.'
            dg = derive(generator, order=kwargs['power'])
            taylor_coeffs = dg([0], mult=False)
            self.generator = [taylor_coeffs.get((k,), 0) for k in range(len(taylor_coeffs))]
        else:
            raise NotImplementedError('Input function not recognized.')
        self.power = len(self.generator) - 1
            
    def action(self, y, **kwargs):
        '''
        Apply the Lie operator g(:x:) to a given lie polynomial y to return the elements
        in the series of g(:x:)y.
        
        Parameters
        ----------
        y: liepoly
            The Lie polynomial on which the Lie operator should be applied on.
            
        Returns
        -------
        list
            List containing g[n]*:x:**n y if g = [g[0], g[1], ...] denotes the underlying generator.
            The list goes on up to the maximal power N determined by self.exponent.ad routine (see
            documentation there).
        '''
        assert hasattr(self, 'generator'), 'No generator set (check self.set_generator).'
        assert hasattr(self, 'exponent'), 'No Lie polynomial in exponent set (check self.set_exponent)'
        ad_action = self.exponent.ad(y, power=self.power)
        assert len(ad_action) > 0
        # N.B. if self.generator[j] = 0, then k_action[j]*self.generator[j] = {}. It will remain in the list below (because 
        # it always holds len(ad_action) > 0 by construction).
        # This is important when calculating the flow later on. In order to check for this consistency, we have added 
        # the above assert statement.
        return [ad_action[j]*self.generator[j] for j in range(len(ad_action))]
        
    def calcOrbits(self, **kwargs):
        '''
        Compute the summands in the series of the Lie operator g(:x:)y, for every requested y,
        by utilizing the routine self.action.
        
        Parameters
        ----------
        components: list, optional
            List of liepoly objects on which the Lie operator g(:x:) should be applied on.
            If nothing specified, then the canonical coordinates are used.
            
        **kwargs
            Optional arguments passed to lie.create_coords.
        '''
        if 'components' in kwargs.keys():
            self.components = kwargs['components']
        else:
            self.components = create_coords(dim=self.exponent.dim, **kwargs) # run over all canonical coordinates.
        self.orbits = [self.action(y) for y in self.components]
        
    def calcFlow(self, t=1, **kwargs):
        '''
        Compute the Lie operators [g(t:x:)]y for every y in self.components.
        
        Parameters
        ----------
        t: float (or e.g. numpy.complex128 array), optional
            Parameter in the exponent at which the Lie operator should be evaluated.
        '''
        if not hasattr(self, 'orbits'):
            raise RuntimeError("No orbits found. Flow calculation requires at least one Lie polynomial to be transported (check self.calcOrbits).")
        # N.B. We multiply with the parameter t on the right-hand side, because if t is e.g. a numpy array and
        # standing on the left, then numpy would put the liepoly classes into its array, something we do not want. 
        # Instead, we want to put the numpy arrays into our liepoly class.
        self.flow = [sum([self.orbits[k][j]*t**j for j in range(len(self.orbits[k]))]) for k in range(len(self.orbits))]
        self.flow_parameter = t
        
    def evaluate(self, z, **kwargs):
        '''
        Evaluate current flow of Lie operator at a specific point.
        
        Parameters
        ----------
        z: subscriptable
            The vector z in the expression (g(:x:)y)(z)
            
        Returns
        -------
        list
            The values (g(:x:)y)(z) for y in self.components.
        '''
        assert hasattr(self, 'flow'), "Flow needs to be calculated first (check self.calcFlow)."
        if 't' in kwargs.keys(): # re-evaluate the flow at the requested flow parameter t.
            self.calcFlow(**kwargs)
            
        if hasattr(z, 'shape') and hasattr(z, 'reshape') and hasattr(self.flow_parameter, 'shape'):
            # If it happens that both self.flow_parameter and z have a shape (e.g. if both are numpy arrays)
            # then we reshape z to be able to broadcast z and self.flow_parameter into a common array.
            # After the application of self.flow, a reshape on the result is performed in order to
            # shift the two first indices to the last, so that the @ operator can be applied as expected
            # (see PEP 456).
            # In this way it is possible to compute n coordinate points for m flow parameters, while
            # keeping the current self.flow_parameter untouched.
            # TODO: may need to check speed for various reshaping options.
            trailing_ones = [1]*len(self.flow_parameter.shape)
            z = z.reshape(*z.shape, *trailing_ones)
            result = np.array([self.flow[k](z) for k in range(len(self.flow))])
            # Now the result has z.shape in its first len(z.shape) indices. We need to bring the first two
            # indices to the rear in order to have an object by which we can apply the conventional matmul operation(s).
            transp_indices = np.roll(np.arange(result.ndim), shift=-2)
            return result.transpose(transp_indices)
        else:
            return [self.flow[k](z) for k in range(len(self.flow))]
        
    def apply(self, z, **kwargs):
        '''
        Apply the current Lie operator g(:x:) onto a single liepoly class or a list of liepoly classes.
        This function is basically intended as a shortcut for the successive call of self.calcOrbits and self.calcFlow.
        
        Parameters
        ----------
        z: liepoly or list of liepoly classes
            The Lie polynomial(s) on which to apply the current Lie operator.
            
        Returns
        -------
        self.flow[0] or self.flow
            Depending on the input, either the liepoly g(:x:)y is returned, or a list g(:x:)y for the given
            liepoly elements y.
        '''
        if z.__class__.__name__ == 'liepoly':
            self.calcOrbits(components=[z], **kwargs)
            self.calcFlow(**kwargs)
            return self.flow[0]
        else:
            self.calcOrbits(components=z, **kwargs)
            self.calcFlow(**kwargs)
            return self.flow
    
    def compose(self, z, **kwargs):
        '''
        Compute the composition of the current Lie operator g(:x:) with another one f(:y:), 
        to return the Lie operator h(:z:) given as
           h(:z:) = g(:x:) f(:y:).
           
        Parameters
        ----------
        z: lieoperator
            The Lie operator z = f(:y:) to be composed with the current Lie operator from the right.
            
        Returns
        -------
        lieoperator
            The resulting Lie operator of the composition.
        '''
        raise NotImplementedError('Composition of Lie operators not yet implemented.')

    def __call__(self, z, **kwargs):
        '''
        Compute the result of the current Lie operator g(:x:), applied to either 
        1) a specific point
        2) another Lie polynomial
        3) another Lie operator
        
        Parameters
        ----------
        z: subscriptable or liepoly or lieoperator
            
        **kwargs
            Optional arguments passed to self.calcFlow. Note that if an additional parameter t is passed, 
            then the respective results for g(t*:x:) are calculated.
            
        Returns
        -------
        list or liepoly or lieoperator
            1) If z is a list, then the values (g(:x:)y)(z) for the current liepoly elements y in self.components
            are returned (see self.evaluate).
            2) If z is a Lie polynomial, then the orbit of g(:x:)z will be computed and the flow returned as 
               liepoly class (see self.apply).
            3) If z is a Lie operator f(:y:), then the Lie operator h(:z:) = g(:x:) f(:y:) is returned (see self.compose).
        '''
        if z.__class__.__name__ == 'liepoly':
            return self.apply(z, **kwargs)
        elif z.__class__.__name__ == self.__class__.__name__:
            return self.compose(z, **kwargs)
        else:
            assert hasattr(z, '__getitem__'), 'Input needs to be subscriptable.'
            if z[0].__class__.__name__ == 'liepoly':
                return self.apply(z, **kwargs)
            else:
                return self.evaluate(z, **kwargs)

    



from itertools import product
from scipy.special import bernoulli
from njet.jet import factorials

import numpy as np

'''
Collection of scripts to deal with the Magnus expansion.

References:
[1]: S. P. Norsett, A. Iserles, H. Z. Munth-Kaas and A. Zanna: Lie Group Methods (2000).
[2]: T. Carlson: Magnus expansion as an approximation tool for ODEs (2005).
[3]: A. Iserles: Magnus expansions and beyond (2008).
[4]: A. Iserles, S. P. Norsett: On the solution of linear differential equations in Lie groups, Phil. Trans. R. Soc. Lond. A, no 357, pp 983 -- 1019 (1999).
'''


class hard_edge:
    '''
    Class to model a polynomial describing a single hard-edge element. Intended to be used
    as values of a liepoly class.
    '''
    
    def __init__(self, values: list, lengths):
        '''
        Parameters
        ----------
        values: list
            A list of floats indicating the s-powers of the hard-edge element. For example,
            [1, 3, -2] has the interpretation: 1 + 3*s - 2*s**2.
        
        lengths: dict
            A dictionary containing the powers of the length within which we want to integrate.
            The dictionary should preferably contain all the powers up to max(1, len(values) - 1).
        '''
        assert len(values) > 0
        self.values = values # a list to keep track of the coefficients of the s-polynomial. 
        self.order = len(self.values)
        self._default_tolerance = 1e-15 # values smaller than this value are considered to be zero. This is used to avoid the proliferation of larger and larger lists containing zeros.
        
        # init lengths
        assert 1 in lengths.keys()
        self._integral_lengths = lengths
    
    def copy(self):
        return self.__class__(values=[v for v in self.values], lengths=self._integral_lengths)
        
    def _convert(self, other, **kwargs):
        '''
        Convert argument to self.__class__.
        '''
        if other.__class__.__name__ != self.__class__.__name__:
            kwargs['lengths'] = kwargs.get('lengths', self._integral_lengths)
            return self.__class__(values=[other], **kwargs)
        else:
            return other
        
    def __mul__(self, other):
        '''
        Multiply two hard-edge coefficients with each other.
        
        Example:
        If [0, 2, 1] are the values of self and [1, 1] are the values of
        the other object, then the result would be:
        (0 + 2*s + 1*s**2)*(1 + 1*s) = 0 + 2*s + 3*s**2 + 1*s**3
        corresponding to [0, 2, 3, 1].
        '''
        if self.__class__.__name__ == other.__class__.__name__:
            assert self._integral_lengths[1] == other._integral_lengths[1] # may be dropped if performance is bad
            vals_mult = [0]*(self.order + other.order)
            max_power_used = 0 # to drop unecessary zeros later on
            for order1 in range(self.order):
                value1 = self.values[order1]
                if abs(value1) <= self._default_tolerance:
                    continue
                for order2 in range(other.order):
                    value2 = other.values[order2]
                    if abs(value2) <= self._default_tolerance:
                        continue
                    vals_mult[order1 + order2] += value1*value2
                    max_power_used = max([max_power_used, order1 + order2])
            result = self.__class__(values=vals_mult[:max_power_used + 1], lengths=self._integral_lengths)
        else:
            result = self.__class__(values=[v*other for v in self.values], lengths=self._integral_lengths)
        return result
    
    def __add__(self, other):
        if self.__class__.__name__ == other.__class__.__name__:
            assert self._integral_lengths[1] == other._integral_lengths[1] # may be dropped if performance is bad
            vals_add = []
            if self.order == other.order:
                max_used = 0 # to remove possible trailing zeros
                for k in range(other.order):
                    value_k = self.values[k] + other.values[k]
                    if abs(value_k) > self._default_tolerance:
                        max_used = k
                    vals_add.append(value_k)
                vals_add = vals_add[:max_used + 1]
            elif self.order > other.order:
                for k in range(other.order):
                    vals_add.append(self.values[k] + other.values[k])
                vals_add += self.values[other.order:]
            else: # self.order < other.order; not that there must be higher-order non-zero values in 'other' which will have no counterpart.
                for k in range(self.order):
                    vals_add.append(self.values[k] + other.values[k])
                vals_add += other.values[self.order:]
            result = self.__class__(values=vals_add, lengths=self._integral_lengths)
        else:
            result = self.__class__(values=[self.values[0] + other] + [v for v in self.values[1:]],
                                   lengths=self._integral_lengths)
        return result
    
    def __neg__(self):
        return self.__class__(values=[-v for v in self.values], lengths=self._integral_lengths)
    
    def __sub__(self, other):
        return self + -other
    
    def __radd__(self, other):
        return self + other
    
    def __rmul__(self, other):
        return self*other
    
    def _integrate(self, constant=0):
        '''
        Integrate the given polynomial from zero to self._integral_lengths[1]
        
        Parameters
        ----------
        constant: float, optional
            An optional integration constant. May be required if the integration
            corresponds to a piecewise integration over a larger region.
            
        Returns
        -------
        hard_edge
            A hard_edge object, containing the coefficients of the integral as their new values.
        '''
        # In order to prevent that we add unecessary zeros to the new values, we may have to shift the maximum index by one.
        # This will be taken into account only at the 'start', where the hard-edges are expected to have no higher-order components.
        n_max = self.order + 1
        if self.order == 1 and abs(self.values[0]) <= self._default_tolerance:
            n_max = 1
            
        # now put the new values and update the integration constant
        power_coeffs = [constant] + [self.values[k - 1]/k for k in range(1, n_max)]
        integral = constant + sum([power_coeffs[mu]*self._integral_lengths.get(mu, self._integral_lengths[1]**mu) for mu in range(1, n_max)])
        return power_coeffs, integral
    
    def integrate(self, **kwargs):
        power_coeffs, integral = self._integrate(**kwargs)
        return self.__class__(values=power_coeffs, lengths=self._integral_lengths), integral # self._integral_lengths may be modified in the original object, by the additional key. This is intended to avoid unecessary calculations.
    
    def __eq__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            # In this case there may be a check of 'self' against a float like zero.
            # we have to return False here, otherwise e.g. liepoly elements containing hard_edge elements may lose some keys
            # (as they will not keep track of keys containing zeros) and eventually drop out.
            # Only under the condition that there were also no integral lengths given we return True.
            return self.order == 1 and abs(self.values[0] - other) <= self._default_tolerance # and len(self._integral_lengths) == 0
        elif self.order != other.order:
            return False
        else: # check the fields, based on successive complexity
            if not all([abs(self.values[k] - other.values[k]) <= self._default_tolerance for k in range(self.order)]):
                return False
            if self._integral_lengths[1] != other._integral_lengths[1]:
                return False
            else:
                return True
    
    def __str__(self):
        return str(self.values)
        
    def _repr_html_(self):
        return f'<samp>{self.__str__()}</samp>'


class hard_edge_chain:
    
    '''
    Class to handle a chain of hard-edge elements, given by piecewise polynomial functions, and their respective integrals.
    '''
    
    def __init__(self, values, **kwargs):
        '''
        Parameters
        ----------
        values: list
            A list of hard edge elements.
        '''
        assert len(values) > 0
        self.values = values # values[k] should be a list of hard_edge or lie-polynomial objects with hard_edge objects as values.
        self._integral = kwargs.get('integral', None)
        
    def copy(self):
        return self.__class__(values=[v.copy() for v in self.values], integral=self._integral)
    
    def __getitem__(self, index):
        return self.values[index]
    
    def __len__(self):
        return len(self.values)
    
    def __mul__(self, other):
        result = []
        if self.__class__.__name__ == other.__class__.__name__:
            assert len(self) == len(other)
            for k in range(len(self)): 
                result.append(self[k]*other[k])
        else:
            for k in range(len(self)):
                result.append(self[k]*other)
        return self.__class__(result, integral=self._integral)
    
    def __rmul__(self, other):
        return self*other
    
    def __add__(self, other):
        '''
        Add two hard_edge_chain's. 
        
        Attention: It is not checked (but assumed) that the respective element names of the hard_edge objects are equal.
        '''
        result = []
        if self.__class__.__name__ == other.__class__.__name__:
            assert len(self) == len(other)
            # the integral lengths of power 1 must be equal; we drop this check for the time being (to improve performance)
            #assert all([self.values[k]._integral_lengths[1] == other.values[k]._integral_lengths[1] for k in range(len(self))])
            for k in range(len(self)):
                result.append(self[k] + other[k])
        else:
            for k in range(len(self)):
                result.append(self[k] + other)
        return self.__class__(values=result, integral=self._integral)
    
    def __radd__(self, other):
        return self + other
    
    def __neg__(self):
        return self.__class__(values=[-v for v in self.values], integral=self._integral)
    
    def __sub__(self, other):
        return self + -other
    
    def __eq__(self, other):
        '''
        Check if the hard_edge_model contains the same values.
        
        Attention: No check regarding self._integral_constant.
        '''
        if self.__class__.__name__ == other.__class__.__name__:
            if len(self.values) != len(other.values):
                return False
            else:
                return all([self.values[k] == other.values[k] for k in range(len(self.values))])
        else:
            return all([self.values[k] == other for k in range(len(self.values))])
        
    def integrate(self):
        '''
        Compute the integral
        
          x
         /
         | h(s) ds
         /
         p0
         
        where p0 = self.positions[0] and h(s) is the Hamiltonian of the hard-edge model.
        
        Returns
        -------
        integral: hard_edge
            A hard_edge object, representing the section-wise integrand of the current hard_edge object.
        '''
        result_values = []
        constant = 0
        for X in self.values:
            IX, constant_last = X.integrate(constant=constant)
            result_values.append(IX)
            constant += constant_last # use individual constants to be added to the integral for the next hard-edge functions
        return self.__class__(values=result_values, integral=constant_last)
    
    def __str__(self):
        out = ''
        for p in self.values:
            out += f'{str(p)} -- '
        return out[:-4]
        
    def _repr_html_(self):
        return f'<samp>{self.__str__()}</samp>'
    
    
class fourier_model:
    
    '''
    Model and compute a specific term in the integral of a tree, in case the original Hamiltonian has been decomposed into a Fourier series.
    '''
    
    def __init__(self, factors=[], exponent={}, sign=1):
        
        self.factors = factors
        self.exponent = exponent # the keys in 'exponent' denote the indices of the s-values, while the items denote the indices of the 'omega'-values. 
        self.sign = sign
        
    def __str__(self):
        if len(self.factors) == 0:
            return '1'
        
        if self.sign == -1:
            out = '-'
        else:
            out = ' '
            
        out += '[ '
        for flist in self.factors:
            out += '('
            for f in flist:
                out += f'f{f} + '
            out = out[:-3]
            out += ')*'
        out = out[:-1]
        expstr = ''
        
        for k, v in self.exponent.items():
            expstr += '('
            for e in v:
                expstr += f'f{e} + '
            expstr = expstr[:-3]
            expstr += ') + '
        expstr = expstr[:-3]
        
        out += f' ]**(-1) * [exp( {expstr} ) - 1]'
        return out
    
    def _repr_html_(self):
        return f'<samp>{self.__str__()}</samp>' 


class tree:
    '''
    A tree according to Refs. [1, 2, 3]
    '''
    
    def __init__(self, *branches, time_power=0, **kwargs):
        '''
        self.integration_scheme
        
        Example:
        Let the tree have the form
          [I([I(H), H]), [I([I(H), H]), H]]
        The function 'H' appears 5 times and each 'I' denotes a specific integration. 
        In addition to this, there is an integration over the entire expression (whose upper bound
        we shall denote by t_-1). 
        By bi-linearity of the Lie brackets, we can move out the interior integrals to obtain:
        
        I_0^{t_j4} [I_0^{t_j1} [I_0^{t_j0} H(t_0), H(t_1)], [I_0^{t_j3} [I_0^{t_j2} H(t_2), H(t_3)], H(t_4)]]
        
        The field self.integration scheme is a list L so that L[k] = jk. For this example:
        
        L = [1, 4, 3, 4, -1]
        
        So that the above integral is given by

        I_0^{t_-1} [I_0^{t_4} [I_0^{t_1} H(t_0), H(t_1)], [I_0^{t_4} [I_0^{t_3} H(t_2), H(t_3)], H(t_4)]]
        
        Parameters
        ----------
        *branches: None or (tree, tree)
            Either none or two trees which define the branches of the given tree.
        
        time_power: int, optional
            Only relevant at initialization if len(branches) == 0. 
            Defines the first order in time in the time-expansion of the operator (Hamiltonian).
        '''
        self._upper_bound_default = -1 # should be smaller than 0
        self.branches = branches
        if len(branches) == 0:
            self.index = 1 # we set the index of the fundamental object to one, to have the easiest way to increase the index for higher-orders.
            self.pivot_branches = [] # Contains the trees listen in e.g. Eq. (2.3), Ref. [3]
            self.factor = 1
            self.time_power = time_power # The power of the first coefficient in the Taylor expansion of the Hamiltonian with respect to time.
            self.integration_bounds = [self._upper_bound_default]
        else:
            assert len(branches) == 2
            self.index = branches[0].index + branches[1].index
            self.pivot_branches = [branches[0]] + branches[1].pivot_branches # Contains the trees listen in e.g. Eq. (2.3), Ref. [3]
            
            ### Keep track of the integration scheme ###
            # copy the integration schemes of the two branches
            bounds1 = [s for s in branches[0].integration_bounds]
            bounds2 = [s for s in branches[1].integration_bounds]
            
            # find their free variables, which are those unique indices with a '-1'.
            free_variable1 = bounds1.index(self._upper_bound_default)
            free_variable2 = bounds2.index(self._upper_bound_default)
            
            # relabel indices of scheme2 to fit in the larger setting
            bounds2 = [s + branches[0].index for s in bounds2]
            bounds2[free_variable2] = self._upper_bound_default # keep the free variable of scheme2
            
            # define integration over the free variable of scheme1 with upper bound of the free variable of scheme2
            bounds1[free_variable1] = free_variable2 + branches[0].index
            
            # put all together to define the integration scheme of the current tree
            self.integration_bounds = bounds1 + bounds2
            
        if 'factors' in kwargs.keys():
            self.set_factor(factors=kwargs['factors'])
            
    def _set_time_power(self, eb=None):
        '''
        Set the time power of the tree.
        
        Parameters
        ----------
        eb: boolean, optional
            self.branches[0] == self.branches[1]; if 'None' given, the check will be done.
        '''
        if eb == None:
            eb = self.branches[0] == self.branches[1]
            
        if eb:
            self.time_power = self.branches[0].time_power + self.branches[1].time_power + 2
        else:
            self.time_power = self.branches[0].time_power + self.branches[1].time_power + 1
            
    def integration_chain(self):
        '''
        Convert the integration bounds of the current tree into a multi-dimensional integral over a simplex. I.e. this routine
        will move the integrals in the example of self.__init__ in front of the entire expression.

        Returns
        -------
        list:
            A list of tuples, denoting the order and bounds of the multi-dimensional integral:
            [(j1, b1), (j2, b2), ..., (jk, bk)] corresponds to

            I_0^{b1} I_0^{b2} ... I_0^{bk} f(t_0, t_1, ..., t_k) dt_j1 ... dt_jk ,
            
            where f(t_0, t_1, ..., t_k) denotes the nested bracket expression of the tree (see self.__init__ for an example).
        '''
        # Input consistency check: In order to be able to move an integral into the front of the nested commutator expression, 
        # it is necessary that for every variable t_k, its corresponding upper bound b_k does not equal one of the
        # preceeding variables.
        assert all([all([j < self.integration_bounds[k] for j in range(k)]) for k in range(len(self.integration_bounds)) if self.integration_bounds[k] != self._upper_bound_default])
        
        # construct the ordering
        default_var = self.integration_bounds.index(self._upper_bound_default)
        level = {default_var: self._upper_bound_default}
        order = [(default_var, self._upper_bound_default)]
        integration_levels = [level]
        while len(level.keys()) > 0:
            level = {k: self.integration_bounds[k] for k in range(len(self.integration_bounds)) if self.integration_bounds[k] in level.keys()}
            order += [e[0] for e in zip(level.items())]
            integration_levels.append(level)
        return order, integration_levels[:-1]
    
    def hard_edge_integral(self, hamiltonian):
        '''
        Compute the nested chain of integrals in case the underlying Hamiltonian is given by a hard-edge model.
        
        This routine is intended to be called on the final tree of the problem.
        
        Parameters
        ----------
        hamiltonian
            A liepoly object having of hard_edge_chain elements as values.

        Returns
        -------
        liepoly
            A liepoly object where every value corresponds to the integral of the given hard_edge_chains over the nested integration.
        '''
        integrands = {k: hamiltonian for k in range(self.index)}
        ic, _ = self.integration_chain()
        for var, bound in ic[::-1]:
            Iham = integrands[var].apply('integrate')
            # bound will be the next element in the integration chain for this specific element.
            # therefore we have to seek out the integrand there and apply the commutator with
            # the hamiltonian at that place
            if bound == self._upper_bound_default:
                break
            assert bound > var # This should be the case by construction of the trees; otherwise the following commutator may have to be reversed.
            integrands[bound] = Iham@integrands[bound]
        return Iham
    
    def fourier_integral_terms(self, consistency_checks=False):
        '''
        If the original t-dependency reads exp(i sum_j (omega_j t_j)), for some variables omega_j, then
        a tree expression can be integrated immediately with respect to the t_j's. This routine will compute
        the resulting factor in front of the exponential, which will depend on the omega_j's. Their indices
        will be returned.
        
        Parameters
        ----------
        consistency_checks: boolean, optional
            If true, perform some consistency checks of the output.

        Returns
        -------
        list
            A list of fourier_model objects, having length 2**(self.index - 1). Each entry A hereby represents 
            one summand of the final integral.
        '''
        integral = [fourier_model(exponent={j: [j] for j in range(self.index)})] # the keys in 'exponent' denote the indices of the s-values, while the items denote the indices of the 'omega'-values.
        if self.index == 1:
            return integral
        
        chain, _ = self.integration_chain()
        for int_op in chain[::-1]:
            variable, bound = int_op
            
            # perform the integration for each summand represented by an entry in 'integral'
            new_integral = []
            for entry in integral:
                new_factors = [e for e in entry.factors] # copy required; otherwise entry['factors'] will be modified unintentionally
                coeffs = entry.exponent[variable]
                new_factors.append(coeffs)

                # lower bound (zero):
                exponent_lower = {k: [c for c in entry.exponent[k]] for k in entry.exponent.keys() if k != variable} # need to reset exponent[k] here as well, otherwise it will get modified unintentionally.
                # upper bound:
                exponent_upper = {a: [e for e in b] for a, b in exponent_lower.items()} # copy required; otherwise exponent_lower gets modified later
                if bound in exponent_upper.keys():
                    exponent_upper[bound] += coeffs
                else:
                    # the final integration
                    assert bound == self._upper_bound_default
                    exponent_upper[bound] = coeffs
                
                # add integration for upper and lower bound to the new set of integrals
                new_integral.append(fourier_model(factors=new_factors, exponent=exponent_upper, sign=entry.sign))
                new_integral.append(fourier_model(factors=new_factors, exponent=exponent_lower, sign=-1*entry.sign))

            integral = new_integral
            
        non_zero_terms = [integral[2*k] for k in range(len(integral)//2)] # by construction, the non-zero exponents (upper bounds) are computed first
        if consistency_checks:
            # consistency checks
            assert len(integral) == 2**self.index # each integration step yields two terms, one for the upper and one for the lower bound.
            zero_terms = [integral[2*k + 1] for k in range(len(integral)//2)] # in the final output, the lower bounds have zero in their exponents.
            assert all([len(e.exponent) == 1 for e in non_zero_terms]) # the non-zero terms must have only self._upper_bound_default as single key.
            assert all([len(e.exponent) == 0 for e in zero_terms]) # verify that the zero_terms are in fact zero in their exponents.
            assert all([e.exponent[self._upper_bound_default] == e.factors[-1] for e in non_zero_terms]) # the last factors must equal the final exponents.
            
        # prepare output; only factors are required for the full information
        return non_zero_terms
            
    def set_factor(self, factors=[]):
        '''
        Set the factor associated with the coefficient of the tree.
        The factor will be given in self.factor.
        
        Attention: Calling this method requires that all the factors in self.pivot_branches have already been determined. There will be no check.
        
        Parameters
        ----------
        b: list, optional
            An optional argument so that the n-th element equals B_n/n!. If such a list is given,
            it must hold len(b) >= len(self.pivot_branches)
        '''
        n = len(self.pivot_branches)
        if len(factors) == 0:
            f0 = bernoulli(n)[-1]/factorials(n)[-1]
        else:
            f0 = factors[n]
            
        self.factor = f0
        for pbranch in self.pivot_branches:
            self.factor *= pbranch.factor
            
    def __eq__(self, other):
        if self.index != other.index:
            return False
        elif self.index == 1: # == other.index
            return True
        else:
            return self.branches[1] == other.branches[1] and self.branches[0] == other.branches[0]
        
    def __str__(self):
        if len(self.branches) == 0:
            return 'H'
        else:
            return f'[I({str(self.branches[0])}), {self.branches[1]}]'
    
    def _repr_html_(self):
        return f'<samp>{self.__str__()}</samp>' 
    
    
def forests(k, time_power=0, **kwargs):
    '''
    Construct a set of trees with respect to a given index, according to Refs. [1, 2, 3].
    
    Parameters
    ----------
    k: int
        The maximal index up to which we want to construct the trees.
        
    time_power: int, optional
        Optional initial order of time, of the first Taylor coefficient of the given operator.
        
    **kwargs
        Optional arguments given to the tree instantiation.
        
    Returns
    -------
    dict
        A dictionary where the key j corresponds to a list containing all trees with index j.
        
    dict
        A dictionary representing the forests of trees for the requested time powers 
        (up to k*(time_power + 1) + 2, see the discussion in this code below).
    '''
    factors = bernoulli(k)/factorials(k)
    tree_groups = {0: [tree(time_power=time_power, **kwargs)]} # Representing the sets T_k in Refs. [1, 2, 3].
    for j in range(1, k + 1):
        # construct the set of trees with respect to index j from trees of indices q < j and p < j so that q + p = j.
        # (note that later on the indices are shifted by one, due to our convention to start with index 1.)
        treesj = []
        for q in range((j - 1)//2 + 1): # We don't have to iterate from 0 up to j - 1, but only half the way: The case j - q - 1 < q is already covered by some q later on (elements of trees_q and trees_p are exchanged and added, if they provide a new unique tree).
            p = j - q - 1
            # N.B. 
            # 1) q <= p
            # 2) Unfortunately there is no immediate relation to the building of trees and the factors: Even if the factor
            # of a specific tree is zero, it may happen that this tree is used later on in a tree with a higher index whose factor
            # is not zero. So for building the forest we need to take all trees into account even if their factors may be zero.
            # For example: Tree nr. 7 in Ref. [2] is zero due to B_3 = 0, but tree nr. 17 is not zero, but build from the tree nr. 7.
            for t1, t2 in product(tree_groups[q], tree_groups[p]):
                branches_equal = t1 == t2 # n.B. if q != p, then this will not go deep into the trees but return False immediately.
                
                t12 = tree(t1, t2, factors=factors, **kwargs)
                # t12.index == j + 1 with p + q == j - 1, so that t12.index == p + q + 2 == t1.index + t2.index
                t12._set_time_power(branches_equal)
                treesj.append(t12)
                    
                if not branches_equal:
                    if t1.index != t2.index: # otherwise it holds q == p and so tree(t2, t1) was already added above.
                        t21 = tree(t2, t1, factors=factors, **kwargs)
                        t21._set_time_power(branches_equal)
                        treesj.append(t21)

        tree_groups[j] = treesj
            
    time_powers = np.unique([t.time_power for tg in tree_groups.values() for t in tg])
    max_power = k*(time_power + 1) + 2 # Trees of index k can only contribute to forests of k + 2 at most (if multiplied by a tree with index 1).
    # Each additional time_power can be extracted out of the brackets and therefore acts as a flat addition to this value. So we have max_power = k + 2 + k*time_power = k*(1 + time_power) + 2. We do not include forests beyond this value.
    forest_groups = {power: [t for tg in tree_groups.values() for t in tg if t.time_power == power] for power in time_powers if power <= max_power}
    # N.B. in Ref. [2] it appears that the term belonging to "F_5" with coeff -1/24 is not correctly assigned. It should be in F_6.
    # Furthermore, from [4] #T_k = (2k)!/k!/(k + 1)!.
        
    return tree_groups, forest_groups


def norsett_iserles(order: int, hamiltonian: hard_edge_chain, time=True, **kwargs):
    '''
    Compute an expansion of the Magnus series, given by Norsett and Iserles (see Ref. [1] in magnus.py) using rooted trees, here in case of hard-edge elements.
    
    Parameters
    ----------
    order: int
        The maximal order to be considered in the expansion. This order corresponds to the accuracy in powers
        of s (the integration variable) of the resulting Hamiltonian.
        
    hamiltonian: hard_edge_chain
        A hard-edge chain of liepoly objects, having values in hard_edge objects. Each liepoly object must have the same amount of keys,
        but their coefficients may differ (or can be set to zero).
        
    time: boolean, optional
        A switch whether to use forests according to the number of involved commutators (time = False) or
        according to the power in s (time = True).
        
    Returns
    -------
    dict
        A dictionary mapping integer values (the orders) to lists, which contain the results of
        integrating the given hard-edge Hamiltonian with respect to the individual trees.
    '''
    forest, tforest = forests(order)
    if time:
        forest_oi = tforest
    else:
        forest_oi = forest
        
    result = {}
    for l in forest_oi.keys():
        if l > order: # the time_power of trees may exceed the maximal index given by the order, therefore we drop these forests.
            continue
        result_l = [] # in general there are several trees for a specific order l
        for tr in forest_oi[l]:
            if tr.factor == 0:
                continue
            I = tr.hard_edge_integral(hamiltonian=hamiltonian)
            if len(I) > 0:
                result_l.append((I, tr.factor))
        result[l] = result_l
    return result



# This script will collect various checks related to Linear Algebra.

import numpy as np
import mpmath as mp

from .matrix import get_package_name

def is_positive_definite(A):
    '''Check if a given matrix A is positive definite.

    Parameters
    ----------
    A: matrix
        The matrix to be checked.
        
    code: str, optional
        The code to be used for the check. Either 'mpmath' or 'numpy' (default).
        
    Returns
    -------
    boolean
        True if matrix is positive definite.
    '''
    code = get_package_name(A)
    
    out = True
    if code == 'mpmath':
        try:
            _ = mp.cholesky(A)
        except:
            out = False
    else:
        try:
            _ = np.linalg.cholesky(A)
        except:
            out = False
    return out
    
def relative_eq(a, b, tol=1e-14, **kwargs):
    '''
    Check if the relative difference between two values is smaller than a given value tol.
    
    Parameters
    ----------
    a: complex
        First parameter
    
    b: complex
        Second parameter
        
    tol: float, optional
        Tolerance below which we consider the relative difference of the two parameters as 
        zero.
        
    Returns
    -------
    boolean
    '''
    if a == 0 and b == 0:
        return True
    else:
        return abs(a - b)/max([abs(a), abs(b)]) < tol
    
def williamson_check(A, S, J, tol=1e-14):
    '''
    Check if a given matrix A and the matrix S diagonalizting A according to the theorem of Williamson
    by S.transpose()*A*S = D actually are satisfying all required conditions of the theorem.
    
    If any condition is violated, the routine raises an AssertionError.
    
    Parameters
    ----------
    A: matrix
        The matrix to be diagonalized.
        
    S: matrix
        The symplectic matrix obtained by Williamson's theorem.
        
    code: str, optional
        The code by which the check should be performed. Either 'mpmath' or 'numpy' (default).
        
    tol: float, optional
        A tolerance by which certain properties (like matrix entries) are considered to be zero (default 1e-14).
    '''
    code = get_package_name(A)

    if code == 'numpy':
        isreal = np.isreal(A).all()
        isposdef = is_positive_definite(A, code=code)
        issymmetric = np.all(A - A.transpose()) == 0
        isevendim = len(A)%2 == 0
        symplecticity = np.linalg.norm(S.transpose()*J*S - J)
        issymplectic = symplecticity < tol
        
        diag = J@S@J@A@J@S.transpose()@J
        offdiag = np.array([[diag[k, l] if k != l else 0 for k in range(len(diag))] for l in range(len(diag))])
        isdiagonal = np.all(np.abs(offdiag) < tol)

    elif code == 'mpmath':
        isreal = mp.norm(A - A.conjugate()) == 0
        isposdef = is_positive_definite(A, code=code)
        issymmetric = all([[(A - A.transpose())[i, j] == 0 for i in range(len(A))] for j in range(len(A))])
        isevendim = len(A)%2 == 0
        symplecticity = mp.norm(S.transpose()@J@S - J)
        issymplectic = symplecticity < tol
        
        diag = J@S@J@A@J@S.transpose()@J
        absoffdiag = np.array([[abs(complex(diag[k, l])) if k != l else 0 for k in range(len(diag))] for l in range(len(diag))])
        isdiagonal = np.all(absoffdiag < tol)
        
    assert isreal, 'Input matrix A not real.'
    assert isposdef, 'Input matrix A not positive definite.'
    assert issymmetric,  'Input matrix A not symmetric.'
    assert isevendim, 'Dimension not even.'
    assert issymplectic, f'Symplecticity not ensured: |S^(tr)@J@S - J| = {symplecticity} >= {tol} (tol)'
    assert isdiagonal, f'Matrix D = S^(-tr)@M@S^(-1) =\n{diag}\nappears not to be diagonal (one entry having abs > {tol} (tol)).'




# This script contains routines which returns the generating series for various functions. 
from njet.jet import factorials

def genexp(power, t=1):
    # The generator of exp(t)
    facts = factorials(power)
    return [t**k/facts[k] for k in range(len(facts))]


# This file collects routines which are focused on representing or generating a matrix and to perform
# fundamental matrix operations in various codes.

import numpy as np
import mpmath as mp

def get_package_name(x):
    '''
    Routine intended to get the package name of a specific object (taken from njet.functions).
    
    Parameters
    ----------
    x: obj
        The object to be examined.
        
    Returns
    -------
    str
        A string denoting the code to be used on the object. 
    '''
    return str(x.__class__.__mro__[0].__module__).split('.')[0]
    
def printmat(M, tol=1e-14):
    # print a matrix (for e.g. debugging reasons)
    M = mp.matrix(M)
    mp.nprint(mp.chop(M, tol))
    
def column_matrix_2_code(M, code, **kwargs):
    # translate a list of column vectors to a numpy or mpmath matrix
    if code == 'numpy':
        return np.array(M).transpose()
    if code == 'mpmath':
        return mp.matrix(M).transpose()

def create_J(dim: int):
    r'''
    Create a 2*dim-square matrix J, represented in form of a list of column vectors,
    corresponding to the standard symplectic block-matrix
    
             /  0   1  \
        J =  |         |
             \ -1   0  /
             
    Parameters
    ----------
    dim: int
        Dimension/2 of the matrix to be constructed.
        
    Returns
    -------
    list
        List of column vectors.
    '''
    dim2 = 2*dim
    J1, J2 = [], []
    for k in range(dim):
        J1.append([0 if i != k + dim else -1 for i in range(dim2)])
        J2.append([0 if i != k else 1 for i in range(dim2)])
    return J1 + J2 


def qpqp2qp(n):
    '''Compute a transformation matrix T by which we can transform a given
    (2n)x(2n) matrix M, represented in (q1, p1, q2, p2, ..., qn, pn)-coordinates, into
    a (q1, q2, ..., qn, p1, p2, ..., pn)-representation via
    M' = T^(-1)*M*T. T will be orthogonal, i.e. T^(-1) = T.transpose().
    
    Parameters
    ----------
    n: int
        number of involved coordinates (i.e. 2*n dimension of phase space)
        
    Returns
    -------
    np.matrix
        Numpy matrix T defining the aforementioned transformation.
    '''
    columns_q, columns_p = [], []
    for k in range(n):
        # The qk are mapped to the positions zj via k->j as follows
        # 1 -> 1, 2 -> 3, 3 -> 5, ..., k -> 2*k - 1. The pk are mapped to the 
        # positions zj via k->j as follows
        # 1 -> 2, 2 -> 4, 3 -> 6, ..., k -> 2*k. The additional "-1" is because
        # in Python the indices are one less than the mathematical notation.
        column_k = np.zeros(2*n)
        column_k[2*(k + 1) - 1 - 1] = 1
        columns_q.append(column_k)

        column_kpn = np.zeros(2*n)
        column_kpn[2*(k + 1) - 1] = 1
        columns_p.append(column_kpn)
        
    q = np.array(columns_q).transpose()
    p = np.array(columns_p).transpose()
    return np.bmat([[q, p]])



def matrix_from_dict(M, code, symmetry: int=0, **kwargs):
    
    '''
    Create matrix from (sparse) dict.
    
    Parameters
    ----------
    M: dict
        The dictionary defining the entries M_ij of the matrix in the form:
        M[(i, j)] = M_ij
        
    n_rows: int, optional
        The number of rows.

    n_cols: int, optional
        The number of columns.
    
    symmetry: int, optional
        If 0, no symmetry is assumed (default). 
        If 1, matrix is assumed to be symmetric. Requires n_rows == n_cols.
        If -1, matrix is assumed to be anti-symmetric. Requires n_rows == n_cols.
        
    code: str
        Requested code passed to 'column_matrix_2_code' routine.
    '''
    assert symmetry in [-1, 0, 1]

    dict_shape = max(M.keys())
    n_rows = kwargs.get('n_rows', dict_shape[0] + 1)
    n_cols = kwargs.get('n_cols', dict_shape[1] + 1)
    
    # create a column-matrix
    if symmetry == 0:
        mat = [[0]*n_rows for k in range(n_cols)]
        for i in range(n_rows):
            for j in range(n_cols):
                mat[j][i] = M.get((i, j), 0)
    else:
        dim = max([n_rows, n_cols])
        mat = [[0]*dim for k in range(dim)]
        for i in range(dim):
            for j in range(i + 1):
                hij = M.get((i, j), 0)
                hji = M.get((j, i), 0)
                if hij != 0 and hji != 0:
                    assert hij == symmetry*hji
                if hij == 0 and hji != 0:
                    hij = symmetry*hji
                # (hij != 0 and hji == 0) or (hij == 0 and hji == 0). 
                mat[j][i] = hij
                mat[i][j] = symmetry*hij
    return column_matrix_2_code(mat, code=code)


class cmat: # TODO: May work on a class to conveniently switch between numpy and mpmath code.
    '''
    Class to model a (sparse) matrix for various codes. 
    '''
    def __init__(self, M, **kwargs):
        # M is assumed to be a dictionary, mapping tuples of indices to values.
        # This means that M can be sparsely defined, but then a 'shape' argument should be provided.
        self.entries = M
        self.rows, self.columns = [], []
        for i, j in self.entries.keys():
            self.rows.append(i)
            self.columns.append(j)
        self.shape = kwargs.get('shape', (max(self.rows) + 1, max(self.columns) + 1))
        
    def tolist(self):
        return [[self.entries.get((i, j), 0) for j in range(self.shape[1])] for i in range(self.shape[0])]
        
    def transpose(self):
        result = {(i, j): self.entries[(j, i)] for j, i in self.entries.keys()}
        return self.__class__(result, shape=(self.shape[1], self.shape[0]))
    
    def conjugate(self):
        result = {tpl: self.entries[tpl].conjugate() for tpl in self.entries.keys()}
        return self.__class__(result, shape=self.shape)
    
    def adjoint(self):
        return self.transpose().conjugate()
    
    def diagonal(self):
        return [self.entries.get((k, k), 0) for k in range(max(self.shape))]
        
    def __matmul__(self, other):
        assert self.shape[1] == other.shape[0]
        result = {}
        common_indices = set(self.columns).intersection(set(other.rows))
        for i in self.rows:
            for k in other.columns:
                result[(i, k)] = sum([self.entries.get((i, j), 0)*other.entries.get((j, k), 0) for j in common_indices])
        return self.__class__(result, shape=(self.shape[0], other.shape[1]))
    
    def __add__(self, other):
        result = {}
        if self.__class__.__name__ != other.__class__.__name__:
            # add value to every entry
            for k in self.entries.keys():
                sum_value = self.entries[k] + other
                if sum_value != 0:
                    result[k] = sum_value
        else:
            assert self.shape == other.shape
            for k in set(self.entries.keys()).union(set(other.entries.keys())):
                sum_value = self.entries.get(k, 0) + other.entries.get(k, 0)
                if sum_value != 0:
                    result[k] = sum_value
        return self.__class__(result, shape=self.shape)
    
    def __radd__(self, other):
        return self + other
    
    def __neg__(self):
        return self.__class__({k: -v for k, v in self.entries.items()}, shape=self.shape)
    
    def __sub__(self, other):
        return self + -other
    
    def __str__(self):
        return repr(self.tolist()) # TMP

    def _repr_html_(self):
        return f'<samp>{self.__str__()}</samp>'


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lieops',
 'lieops..ipynb_checkpoints',
 'lieops.linalg',
 'lieops.linalg..ipynb_checkpoints',
 'lieops.ops',
 'lieops.ops..ipynb_checkpoints']

package_data = \
{'': ['*']}

install_requires = \
['mpmath>=1.2.1,<2.0.0', 'njet>=0.2.2,<0.3.0', 'sympy>=1.9,<2.0']

setup_kwargs = {
    'name': 'lieops',
    'version': '0.1.0',
    'description': 'Lie operator tools.',
    'long_description': None,
    'author': 'Malte Titze',
    'author_email': 'mtitze@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://lieops.readthedocs.io/en/latest/index.html',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
